#! /usr/bin/env python

import sys
import os
import tempfile
import re
import string
import traceback
import commands
import JimConfigManager 
from url_read_config_value import getAttrFromXMLFile
from xml_meta_configurator_lib.XMLDocDefaultTemplate import XMLDocDefualtTemplate


symLinksDirName = "Symlinks"

class UsageException(Exception):
    def __init__(self,strError=""):
        self._strError = strError
    def __str__(self):
        return self._strError
## ==== end of class UsageError ==== ##

class SyncFailedException(Exception):
    def __init__(self,strError=""):
        self._strError = strError
    def __str__(self):
        return self._strError
## ==== end of class SyncFailedException ==== ##      

def printInfo(stringToPrint):
    print "sync_config.py:  INFO   : %s" % stringToPrint
def printError(stringToPrint):
    print "sync_config.py:  ERROR  : %s" % stringToPrint
def printWarning(stringToPrint):
    print "sync_config.py: WARNING : %s" % stringToPrint
# ==== end of printers

sep = "********************************************************************"
class SyncConfig:
    def __init__(self,jimUpsDb,prdName,prdVersionNew,prdVersionOld,prdQualifiers):
        self._jimUpsDb      = jimUpsDb
	self._prdName       = prdName
	self._prdVersionNew = prdVersionNew
	self._prdVersionOld = prdVersionOld
	self._prdQualifiers = prdQualifiers
	
        if (not self._prdVersionOld) or (self._prdVersionOld == "NULL"):
	    printInfo("Getting existing current version from File system")
	    self._prdVersionOld = self.getprdVersionOld()
	
        self._prdStringOld = "%s %s" %  (self._prdName,self._prdVersionOld)
        self._prdStringNew =  "%s %s" % (self._prdName,self._prdVersionNew)
	
	if prdQualifiers and len(string.strip(prdQualifiers)):
            self._prdStringNew = "%s -q '%s'"  % (self._prdStringNew, self._prdQualifiers)
	    self._prdStringOld = "%s -q '%s'"  % (self._prdStringOld, self._prdQualifiers)

        printInfo("Old product version: %s" % self._prdVersionOld)
        printInfo("New product version: %s" % self._prdVersionNew)	
    # ==== end of init

    def getActualPath(self,link):
        # PM: This is a patch to get actual path from symlinks
        # This patch is python version specific
        try:
            return os.path.realpath(link)
        except AttributeError:
            return os.readlink(link)
    # ==== end of getActualPath

    def getSymlinksDir(self):
        return os.path.abspath("%s/%s/Symlinks" % (self._jimUpsDb,self._prdName)) 
    # ==== end of getSymlinksDir

    def getprdVersionOld(self):
        current = "%s/current" % self.getSymlinksDir()
	if not os.path.exists(current):
	    raise SyncFailedException("Cannot read old configuration. Invalid path %s" % current)
        oldVersion = os.path.basename( os.path.abspath(self.getActualPath(current)) )
        if oldVersion == "NULL":
            oldVersion = os.path.basename( os.path.abspath(self.getActualPath(current+"/..")) )
        return oldVersion
    # ==== end of getprdVersionOld

    def synchronize(self):
        oldTemplFile = self.getOldTemplFile()
        newTemplFile = self.getNewTemplFile()
        oldTemplVer  = self.getVersionFromTempl(oldTemplFile)
        newTemplVer  = self.getVersionFromTempl(newTemplFile)
	printInfo("Old template version: %s" % oldTemplVer)
	printInfo("New template version: %s" % newTemplVer)
	
	if not oldTemplVer == newTemplVer:
            raise SyncFailedException("The configuration schema has changed \
 in the new version of %s" % self._prdName) 
 
        self.synchronizeFS() 
	self.synchronizeDB()
        printInfo(sep)
	printInfo("Fininshed synchronization successfuly")
        printInfo(sep)
	return 0
    # ==== end of synchronize
    
    def synchronizeFS(self):
        printInfo("Synchronizing FS configuration")
	oldConfigDir = "%s/%s/%s" % (self._jimUpsDb,self._prdName,self._prdVersionOld) 
        newConfigDir = "%s/%s/%s" % (self._jimUpsDb,self._prdName,self._prdVersionNew)    
        
        if not os.path.exists(newConfigDir) :
	    printInfo("Creating directory: %s" % newConfigDir)
	    cmd = "mkdir -p %s" % newConfigDir
	    status,stdouterr = commands.getstatusoutput(cmd)
	    if not status == 0:
	        raise SyncFailedException("Failed to create directry %s - Reason: %s" \
		% (newConfigDir,stdouterr))
        # end if (both)             

        #PM:  Do a blind copy of the contents from old current to new
        # This is not an elegent solution but works

        cmd = "cp -R %s/* %s" % (oldConfigDir,newConfigDir) 
        status,stdouterr = commands.getstatusoutput(cmd)
        if not status == 0:
	    raise SyncFailedException("Failed to copy files - Reason: %s" % (stdouterr)) 
        
	printInfo(sep)
        printInfo("Your configuration was copied from: %s" % oldConfigDir)
        printInfo("Your configuration was copied to  : %s" % newConfigDir)
        printInfo(sep)
    # ==== end of synchronizeFS
    
    def synchronizeDB(self):
        printInfo("Synchronizing DB configuration") 
        try :
            jcmOld = JimConfigManager.JimConfigManager.JimConfigManager(
                     self._prdName,
                     self._prdVersionOld,
                     self._prdQualifiers)
	    configStr = jcmOld.getConfigString()
            if configStr and len(string.strip(configStr)):
                printInfo("Old configuration: \n%s" % configStr)
                jcm_new = JimConfigManager.JimConfigManager.JimConfigManager(
                      self._prdName,
                      self._prdVersionNew,
                      self._prdQualifiers)
                jcm_new.storeConfigString(configStr)
	    else :
	        raise Exception("Old db config is blank")
	except Exception,ex:
	    printWarning(ex)
	    printWarning("DB Synchronization failed (Not fatal)")
	    return	
    # ==== end of synchronizeDB

    def xpath2opath(self,xpath):
        # SJ: carrying over from previous version of sync_config.py
        #Strip off leading and trailing '/'
        if xpath[0] == '/': xpath=xpath[1:]
        if xpath[-1] == '/': xpath=xpath[0:-1]

        #XML elements may have '-' but PY objects cannot
        result = string.replace(xpath,'-','_')
        result = string.replace(result, '/', '[0].')
        return result
    # ==== end of xpath2opath
 
    def getVersionFromTempl(self,file_path):
        try:
            xpath   = "%s_configuration/interview_schema/version" % self._prdName 
            fileStr = open(file_path,'rt').read()
            from xml_common_lib.PyXmlObject import PyXmlObject
            obj = PyXmlObject(fileStr)
            expr = 'obj.%s' % self.xpath2opath(xpath)
            return eval(expr)
        except Exception,ex:
            raise SyncFailedException("Cannot get templ version: %s" % ex)
    # ==== end of getVersionFromTempl

    def getNewTemplFile(self):
        # Read the version of new template
	if not os.environ.has_key('UPS_PROD_DIR'):
	    raise Exception("Cannot proceed without UPS_PROD_DIR")
        releaseDir   = os.environ['UPS_PROD_DIR']
        templateFile = "%s/etc/%s_config_template.xml" % (releaseDir,self._prdName)
        if not os.access(templateFile,os.R_OK):
	    raise SyncFailedException("No such file: %s" % templateFile)
	printInfo("New template file: %s" % templateFile)
	return templateFile
    # ==== end of getNewTemplFile 

    def getOldTemplFile(self):
        #temp = tempfile.mktemp()
        setupStr = "(source %s/setups.sh; setup %s;echo $%s) 2>/dev/null" %\
        (os.environ['SETUPS_DIR'],self._prdStringOld,string.upper(self._prdName + '_dir'))
	status,releaseDir = commands.getstatusoutput(setupStr)

        if not status == 0:
	    raise SyncFailedException("Failed to get old template version: %s" % releaseDir)
         
        templateFile = "%s/etc/%s_config_template.xml" % (releaseDir,self._prdName)
        if not os.access(templateFile,os.R_OK):
	    raise SyncFailedException("No such file: %s" % templateFile)
	
	printInfo("Old template file: %s" % templateFile)
	return templateFile
    # ==== end of getOldTemplFile  
# ==== end of class SyncConfig ==== #


def doGoodPrint(prdName):
    printInfo("Success!! . Please check configuration for product %s" % prdName)
    printInfo("To change configuration run \"ups tailor %s\"" % prdName)
# ==== end of doGoodPrint    

def doFatalPrint(ex,prdName):
    printError(sep)
    printError("Cannot synchronize or create default config because of the\
 following fatal error:")
    printError(ex)
    printError("PLEASE RUN ups tailor %s" % prdName)
    printError(sep)	
    traceback.print_exc()
# ==== end of doFatalPrint


def createDefaultConfig(jimUpsDb,prdName,prdVersionNew,prdQualifiers):
    try :
        prdString = "%s %s" % (prdName, prdVersionNew)
	if prdQualifiers and len(string.strip(prdQualifiers)):
            prdString = "%s -q '%s'"  % (prdString, prdQualifiers)

        if not os.environ.has_key('UPS_PROD_DIR'):
	    raise Exception("Cannot proceed without UPS_PROD_DIR")
        
        templFile = "%s/etc/%s_config_template.xml" \
	         % (os.environ['UPS_PROD_DIR'],prdName) 
        printInfo("Creating default configuration from template %s" % templFile)
	print ""
        templ = XMLDocDefualtTemplate(templFile)
        output = templ.instantiate()
	print ""
	printInfo("Default configuration looks like\n%s" % output)
        
        
	configDir = "%s/%s/%s" % (jimUpsDb,prdName,prdVersionNew)
        if not os.path.exists(configDir):
  	    status,stdouterr = commands.getstatusoutput("mkdir -p %s" % configDir)
	    if status :
	        raise Exception("Could not create directory %s Reason:%s" % (configDir,stdouterr))
	
	configFile = os.path.join(configDir,"config")
        printInfo("Writing file %s" % configFile) 
	f = open(configFile,'w') 
	f.writelines(output)
	f.close()

        jcm = JimConfigManager.JimConfigManager.JimConfigManager(prdName,prdVersionNew,prdQualifiers)	
        jcm.setMode('DBFS')
        jcm.storeConfigString(output)

        # and the FS too
	printInfo("Stored Default configuration successfully")
	doGoodPrint(prdName)
        
        printInfo("\"Acting\" on the configuration, if needed\n%s" % output)
        act_cmd="[ -e %s/act.py ] && %s/act.py " % (os.environ['UPS_UPS_DIR'],os.environ['UPS_UPS_DIR'])
        printInfo("Exectuing \"%s\"\n" % act_cmd)
        (status, stdouterr) = commands.getstatusoutput(act_cmd)
        print stdouterr
        if status != 0:
          raise Exception("Could not \"act\" on product configuration")
	return 0
    except Exception,ex:
        doFatalPrint(ex,prdName) 
        return 1 

# ==== end of createDefaultConfig

def readArgs(argv, option):
    try:
        for i in range(1,len(argv),1):
            if argv[i] == option:
                return argv[i+1]
        # end for  
        return None 
    except:
        return None 
# ==== end of readArgs	

def usage():
    printError("USAGE: sync_config.py <package name> -n <new version>\
 [-o old version] [-q qualifiers]")
    return 1
# ==== end of usage

def main(args):
    try :
        # PARSE ARGUMENTS 
        if (len(sys.argv) < 2) or (len(sys.argv) > 6):
            raise UsageException("Invalid number of arguments")

        jimUpsDb = os.environ.get('JIM_UPS_DB')
        if not jimUpsDb :
	    raise UsageException('No JIM_UPS_DB in environment. Setup jim_config first')

        prdName = sys.argv[1]
        prdVersionNew = readArgs(sys.argv, '-n') 
        if not prdVersionNew:
	    raise UsageException('Please specify new product version with "-n" option')

        prdVersionOld = readArgs(sys.argv, '-o')
        prdQualifiers = readArgs(sys.argv, '-q')	    

        sc = SyncConfig(jimUpsDb,prdName,prdVersionNew,prdVersionOld,prdQualifiers)
        sc.synchronize()

    except UsageException,ex:
        printError(ex)
        return usage()

    except SyncFailedException,ex :
        printWarning(sep)
        printWarning("Cannot continue synchronization for the following reason")
	printWarning(ex)
        printWarning("Synchronization failed")
	printWarning(sep)
	printWarning("Creating default configuration")
	return(createDefaultConfig(jimUpsDb,prdName,prdVersionNew,prdQualifiers))

    except Exception,ex:
        doFatalPrint(ex,prdName) 
        return 1 

    doGoodPrint(prdName)
    return 0

if __name__ == "__main__":
    sys.exit(main(sys.argv))
