"""
This python script makes managing the configuration in the database easy
"""

__author__="Parag Mhashilkar <parag@fnal.gov>"
__date__= "May 05 2003"
__version__="1.0"

# $source: BaseDBConfigManager.py,v$

################################################################################
# Project   : SAM
# Package : jim_config
#
# This work is part of a development project, called JIM, which consists of a
# number of coordinated packages each named jim_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import os, string, sys, re, traceback
import xmldb_client
import xml_common_lib
from BaseConfigManager import BaseConfigManager
import xmldb_client.xmlxpath_tools


def create_collections(db, collections):
   tokens = string.split(collections, '/')

   currContext = '/db'

   for token in tokens:
      if token == "" or token == "db":
         pass
      else:
         currCollection = currContext + '/' + token
         if not db.IsCollection(currCollection):
            db.CreateCollection(currContext, token)
         currContext = currContext + '/' + token
   return 0
# create_collections


################################################################################
# Class: BaseDBConfigManager
################################################################################
 

class BaseDBConfigManager(BaseConfigManager):
   """This utility class manages the access to Db"""

   def __init__(self, name):
      """
         Contructor to create a BaseDBConfigManager

         name    - Name of the package
      """

      BaseConfigManager.__init__(self, name)
   # __init__

   
   def printConfigDetails(self):
      self.printDebug('-----------------------------------------------------------')
      self.printDebug('My Name               : ' + self.name)
      self.printDebug('My Context            : ' + self.context)
      self.printDebug('Name of configuration : ' + self.configDoc)
      self.printDebug('-----------------------------------------------------------')
   # printConfigDetails


   def getConfigAttr(self, xpath, attr):
      """
         Get a single configuation attribute from the database

         xpath - Xpath to the attribute
         attr - Name of the attribute
         returns a single configuration attribute value in the form of string
      """

      try:
         self.printDebug('Retrieving the configuration attribute '\
                     + xpath + '/' + attr + ' from XML database ...\n')
         dbConn = xmldb_client.xmldb.xmldb()
         return xmldb_client.xmlxpath_tools.readValue(dbConn, self.context,
                                                      xpath, attr)         
      except:
         if self.getDebugMode(): traceback.print_exc()
         raise self.ConfigManagerError('error retrieving config attr from DB')
   # getConfigAttr


   def getConfigString(self):
      """
         Get the configuration from database
 
         return configuration in the form of string else None
      """
      try:
         self.printDebug('Retrieving the configuration '\
                     + self.context + '/' + self.configDoc \
                     + ' from XML database ...\n')
         dbConn = xmldb_client.xmldb.xmldb()
         return dbConn.GetDocument(self.context, self.configDoc)
      except:
         if self.getDebugMode(): traceback.print_exc()
         raise self.ConfigManagerError('error retrieving the config from DB')
   # getConfigString


   def storeConfigString(self, config_str):
      """
         Store the configuration in the form of xml string to the database

         config_str - Configuration in the form of xml string
      """

      try:
         dbConn = xmldb_client.xmldb.xmldb()
         create_collections(dbConn, self.context)
         dbConn.InsertDocument(self.context, self.configDoc, config_str)
         self.printDebug('Your configuration was successfully stored in '\
                     + self.context + '/' + self.configDoc + ' ...\n')
         return 0
      except:
         raise self.ConfigManagerError('error storing the configuration to DB')
   # storeConfigString


# Class BaseDBConfigManager
