#=============================================================================
#
#Project   : JIM
#Package : jim_config
#$Id: jim_sam_storage_negotiator.py,v 1.15 2007/05/10 17:20:22 abaranov Exp $
#
#This work is part of a development project, called JIM, which consists of a
#number of coordinated packages each named sam_xxxx .
#
#Notice of authorship, copyright status,  and terms and conditions, should
#the software eventually become available for use outside Fermilab, can be
#found in the README and LICENCE files in the top level directory of the main
#sam package.
#
#==============================================================================

from JimConfigManager import JimConfigManager

#make sure old storage configuration is support by using GW_SANDBOX_WORK_DIR and local_storage config element
# in the site config

import os

DEBUG=0
HOSTNAME=None

testJobConfigs = { 1: """
<jim_job_managers_configuration>
  <interview_schema version="2_1" />
  <ups setup_script="/local/ups/etc/setups.sh" />
  <experiment experiment_name="d0">
    <sam_prd_qualifier sam_prd_qualifier="prd" />
  </experiment>

  <dzero_monte_carlo>
    <events number_per_output_file="250" />
    <accelerator interval="300" bootstrap_time="300" />
    <local_data_buffer>
      <output_storage name="durable_storage" />
    </local_data_buffer>
  </dzero_monte_carlo>

<dzero_merge>
    <accelerator interval="90" bootstrap_time="180" />
    <local_data_buffer>
      <input_storage name="dodge_binary_storage">
        <prot_fcp queueName="dodge_binary_storage_compat_queue" />
      </input_storage>

    </local_data_buffer>
  </dzero_merge>

  <dzero_reconstruction>
    <accelerator interval="300" bootstrap_time="180" />
    <local_data_buffer>
      <input_storage name="affinity_binary_storage" >
         <prot_fcp queueName="affinity_binary_group" />
      </input_storage>
      <output_storage name="durable_storage" />
    </local_data_buffer>
  </dzero_reconstruction>

  <dzero_reco_merge>
    <accelerator interval="180" bootstrap_time="180" />
    <local_data_buffer>
      <input_storage name="binary_storage" />
    </local_data_buffer>
  </dzero_reco_merge>

  <fss_stager>
    <local_data_buffer>
      <output_storage name="grid1_fss_buffer">
        <prot_fcp queueName="grid1_fss_buffer" />
      </output_storage>
    </local_data_buffer>
  </fss_stager>

  <binary_fetch>
    <local_data_buffer>
      <input_storage name="local_storage" />
    </local_data_buffer>
  </binary_fetch>

  <local_tmp_area local_tmp_area="/local/jim/jim_tmp" />
  <polling_interval grid_update_interval="3600" xml_update_interval="3600" />
</jim_job_managers_configuration>
""",
2 : """
"""
}

testSiteConfigs = { 1: """
<site_configuration>
  <schema version="1_1" />
   <site name="Wuppertal2LCG" />
  <cluster architecture="Linux+2.4" name="lcg_prodcution">
    <gatekeeper location="grid-lcg.physik.uni-wuppertal.de:2119">
      <jobmanager name="jobmanager-samgrid" max_submit_jobs="10" max_jobs="200">
        <station name="ccin2p3-grid1" experiment="d0" universe="prd" />
      </jobmanager>
    </gatekeeper>
  </cluster>
<local_storage path="/d0cache/grid1/durable_loc" node="ccsvli02.in2p3.fr" />

    <local_data_buffer>
    <output_storage name="grid1_fss_buffer" location_selector_algorithm="random" location_selector_pattern="dodge.hep.man.ac.uk:/data1/sam/buffer|dodge.hep.man.ac.uk:/data/sam/buffer" />
    <output_storage name="durable_storage" location_selector_algorithm="random" location_selector_pattern="dodge.hep.man.ac.uk:/data1/sam/durable|dodge.hep.man.ac.uk:/data/sam/durable" />
    <input_storage name="binary_storage" location_selector_algorithm="random" location_selector_pattern="ccsvli02.*d0cache2|caterpillar" />
    <input_storage name="local_storage" location_selector_algorithm="local" location_selector_pattern="ccsvli02.*d0cache2|caterpillar" />
    <input_storage name="dodge_binary_storage" location_selector_algorithm="random" location_selector_pattern="caterpillar" />
    <input_storage name="affinity_binary_storage" location_selector_algorithm="affinity" location_selector_pattern="ccsvli02.*d0cache2||sammy::caterpillar" >
       <fcp_queue name="cms_unmerged_tmbs"/>
       <fcp_queue name="row_unmreged_tmbs"/>
    </input_storage>
    <fcp_queue_group name="affinity_binary_group" queue_selector_algorithm="affinity" queue_selector_pattern="sammy::cms_unmerged_tmbs||row_unmreged_tmbs" />

  </local_data_buffer>
</site_configuration>
""",
2 : """
"""
}

jcm  = JimConfigManager.JimConfigManager("jim_job_managers")

def getHostName():
  import socket
  fqn = socket.gethostbyaddr(socket.gethostname())
  hostName = fqn[0]
  if ( HOSTNAME ):
    hostName = HOSTNAME
  return hostName

def get_queue_parameters(local_data_buffer,group_name):

  try:
    fcp_queue_group_defs = local_data_buffer.fcp_queue_group

    queue_selector_algorithm = None
    queue_selector_pattern = None
    for fcp_queue_group_def in fcp_queue_group_defs:
       if ( str(fcp_queue_group_def.name) == group_name ):
          queue_selector_algorithm = fcp_queue_group_def.queue_selector_algorithm
          queue_selector_pattern = fcp_queue_group_def.queue_selector_pattern
          break

    return ( queue_selector_algorithm , queue_selector_pattern )

  except:
    print "fcp qroup rule is not configured"
    raise Exception("Queue group parameters are not found:"+group_name)
  
def get_storage_parameters(storage_tag,appname):

  try:
   
   global jcm

   if ( DEBUG ):
      from xml_common_lib import PyXmlObject
      application_config = PyXmlObject.PyXmlObject(testJobConfigs[DEBUG]).jim_job_managers_configuration[0].__dict__[appname][0]
   else:
      application_config = jcm.getConfigPyXmlObject().jim_job_managers_configuration[0].__dict__[appname][0]

   storage = application_config.local_data_buffer[0].__dict__[storage_tag][0]

   storage_name = str(storage.name)
   
   try:
     storage_fcp_q = str(storage.prot_fcp[0].queueName)
   except:
     storage_fcp_q = None
 
   if ( DEBUG ):
      local_data_buffer = PyXmlObject.PyXmlObject(testSiteConfigs[DEBUG]).site_configuration[0].local_data_buffer[0]
   else:
      local_data_buffer = jcm.getSiteConfigPyXmlObject().site_configuration[0].local_data_buffer[0]
    

   storages_def = local_data_buffer.__dict__[storage_tag]

   location_selector_pattern = None
   location_selector_algorithm = None
   supported_queues = None

   for storage_def in storages_def:
      if ( str(storage_def.name) == storage_name ):
         location_selector_pattern = str(storage_def.location_selector_pattern)
         location_selector_algorithm = str(storage_def.location_selector_algorithm)
         try:
           supported_queues = map(lambda x: str(x.name) , storage_def.fcp_queue)
         except Exception,ex:
           print "No queues are supported",str(ex)
         break

   queue_selector_algorithm = None
   queue_selector_pattern = None

   if storage_fcp_q and supported_queues:
      ( queue_selector_algorithm , queue_selector_pattern ) = get_queue_parameters(local_data_buffer,storage_fcp_q)

   return ( storage_name , location_selector_pattern , location_selector_algorithm, storage_fcp_q, queue_selector_pattern, queue_selector_algorithm ,supported_queues )

  except Exception,ex:
    print str(ex)
    raise Exception("storage "+storage_tag+" is not configured for "+appname+":" +str(ex))

   
def select_location(nodes,location_selector_algorithm):

   if ( len(nodes) > 1 ):
       if ( location_selector_algorithm != "random" ): 
          import socket
          fqn = socket.gethostbyaddr(socket.gethostname())
          for node in nodes:
             if ( node == fqn[0] ):
                return node
       import random
       node = nodes[random.randint(0, len(nodes)-1)]
   else:
       node = nodes[0] 

   return node

class QueueSelector:
  def __init__(self,supported_queues):
     self._supportedQueues = supported_queues

  def __call__(self,queue_selection_pattern):
     import re
     r = re.compile(queue_selection_pattern)
     if ( not r ):
       raise Exception("Failed to compile configured queue selection re"+str(queue_selection_pattern))

     queues = []
     for q in self._supportedQueues:
       if ( r.match(q) ):
          queues.append(q)
     
     if (  len(queues) == 0 ):
       raise Exception("Failed to select any queue from given list:"+str(self._supportedQueues)+" using regexp "+queue_selection_pattern)

     return queues
   
class InputStorageProvider:
  def __init__(self,stationName):
     self._stationName = stationName

  def __call__(self,location_selector_pattern):
     import random
     tempFileName="/tmp/storageNegotiator2.err"+str(random.randint(0,100000))
     # dump station select nodes
     cmd = """sam dump station --station=%s --disks 2>>%s | grep -v INACTIVE | grep -v DISALLOWED | awk 'BEGIN { FS=" " } /disk [0-9][0-9]* (%s)/{ nodeString = $3; sub(",","",nodeString); while( sub("://","_SEPARATOR_",nodeString) ); split(nodeString,nodeStringArray,":"); node = nodeStringArray[1]; while(sub("_SEPARATOR_", "://", node )); print node; }' 2>>%s""" % ( self._stationName, tempFileName , location_selector_pattern ,tempFileName)
     import commands
     (status , nodes ) = commands.getstatusoutput(cmd)
 
     import os
     os.system("echo Std err:  >&2; cat "+tempFileName+" >&2 ; rm -rf "+tempFileName)
 
     if ( status != 0 or len(nodes) == 0 ):
        import sys
        sys.stderr.write("Command output "+nodes +" , exit status "+str(status))
        raise Exception("Failed to dump station with %s or match with %s pattern is not found" % (cmd,location_selector_pattern) )
 
     nodes = nodes.split("\n")
     return nodes

class OutputStorageProvider:
  def __init__(self,station):
    pass
  def __call__(self,crit):
     return crit.split("|")


def pickRandom(nodes):
   if (  len(nodes) > 1 ):
      import random
      node = nodes[random.randint(0, len(nodes)-1)]
   else: 
      node = nodes[0]
   return node

def RandomStorageSelector(location_selector_pattern,storage_provider):
    nodes = storage_provider(location_selector_pattern)
    return pickRandom(nodes)

def LocalStorageSelector(location_selector_pattern,storage_provider):
    nodes = storage_provider(location_selector_pattern)

    hostName = getHostName()

    for node in nodes:
      if ( node == hostName ):
         return node
    return pickRandom(nodes)

def AffinityStorageSelector(location_selector_pattern,storage_provider):
    # pre parse the pattern to extract affinity definitions <affinity regexp>::<storage selection regexp>|...
    reAlternatives = location_selector_pattern.split("||")

    hostName = getHostName()

    nodes = []
    defaultNodes = []

    for reAlternative in reAlternatives:
       affinityDef = reAlternative.split("::")
       if ( len(affinityDef) > 1 ):
          import re
          r = re.compile(affinityDef[0])
          if ( r.match(hostName) ):
              nodes += storage_provider(affinityDef[1])
       else:
          defaultNodes += storage_provider(affinityDef[0])

    if ( len(nodes) == 0 ):
       nodes = defaultNodes

    if ( len(nodes) > 0 ):
       return pickRandom(nodes)

    raise Exception("Affinity not found for "+location_selector_pattern + " at "+hostName)

storageSelectionAlgo = { "local" : LocalStorageSelector, "random" : RandomStorageSelector, "affinity" : AffinityStorageSelector }
queueSelectionAlgo = { "random" : RandomStorageSelector, "affinity" : AffinityStorageSelector }

def jim_sam_incoming_storage_negotiator(stationName,appname):

   try:
      (storage_name , location_selector_pattern , location_selector_algorithm , storage_fcp_q , queue_selector_pattern, queue_selector_algorithm ,supported_queues ) = get_storage_parameters("input_storage",appname)
   except:
      location_selector_pattern = ".*"
      location_selector_algorithm = "random"
      storage_fcp_q = None
      supported_queues = None

   if supported_queues and storage_fcp_q:
     global queueSelectionAlgo
     storage_fcp_q = queueSelectionAlgo[queue_selector_algorithm](queue_selector_pattern,QueueSelector(supported_queues))

   global storageSelectionAlgo
   node = storageSelectionAlgo[location_selector_algorithm](location_selector_pattern,InputStorageProvider(stationName))

   if ( not ( node and len(node) ) ):
     raise Exception("Failed to select location from "+nodes+" using "+location_selector_algorithm)

   return ( node, storage_fcp_q )
 

def getLocalStorage():
  jcm = JimConfigManager.JimConfigManager()
  path = jcm.getSiteConfigAttr('/site_configuration/local_storage', 'path')
  node = jcm.getSiteConfigAttr('/site_configuration/local_storage', 'node')
  return node+":"+path


def jim_sam_outgoing_storage_negotiator(stationName,appname):

 try:
  (storage_name , location_selector_pattern , location_selector_algorithm , storage_fcp_q , queue_selector_pattern, queue_selector_algorithm ,supported_queues ) = get_storage_parameters("output_storage",appname)

  if supported_queues and storage_fcp_q:
     global queueSelectionAlgo
     storage_fcp_q = queueSelectionAlgo[queue_selector_algorithm](queue_selector_pattern,QueueSelector(supported_queues))

  global storageSelectionAlgo
  node = storageSelectionAlgo[location_selector_algorithm](location_selector_pattern,OutputStorageProvider(stationName))

  if ( len(node) == 0 ):
     raise Exception("0 length location, Failed to select output storage location with pattern %s" % location_selector_pattern)

  return ( node , storage_fcp_q )
 except Exception,ex:
   print str(ex)
   if ( appname == "fss_stager" ):
      return ( os.environ["GW_SANDBOX_WORK_DIR"], None ) 
   return ( getLocalStorage() , None )
  
def asser(a,b):
  if ( a != b ):
     raise Exception("Assert failed."+str(a)+" "+str(b))
if __name__ == "__main__":
  asser ( jim_sam_incoming_storage_negotiator("ccin2p3-grid1","dzero_merge") , ('caterpillar.hep.man.ac.uk', 'dodge_binary_storage_compat_queue' ) )
  asser ( jim_sam_incoming_storage_negotiator("ccin2p3-grid1","dzero_reconstruction"), ('caterpillar.hep.man.ac.uk', 'cms_unmerged_tmbs') )
  asser ( jim_sam_incoming_storage_negotiator("ccin2p3-grid1","binary_fetch"), ('ccsvli02.in2p3.fr', None) )

  try:
    asser ( jim_sam_outgoing_storage_negotiator("ccin2p3-grid1","fss_stager"), ('dodge.hep.man.ac.uk:/data1/sam/buffer', 'grid1_fss_buffer') )
  except:
    try:
      asser ( jim_sam_outgoing_storage_negotiator("ccin2p3-grid1","fss_stager"), ('dodge.hep.man.ac.uk:/data1/sam/buffer', 'grid1_fss_buffer') )
    except:
       asser ( jim_sam_outgoing_storage_negotiator("ccin2p3-grid1","fss_stager"), ('dodge.hep.man.ac.uk:/data1/sam/buffer', 'grid1_fss_buffer') )

