import os
import sys
import time
import random
import traceback
import string
import commands
import re

import SAM

from Sam import sam
import SAMDbServer

from MetadataHelper import _metaAttributes
from  MetadataHelper import _samMDCache


availableTierPrefixes = { "unmerged-filtered-raw" : "zeroB" , "unmerged-thumbnail" : "recoT" , "thumbnail" : "recoT" }

def sortInputFiles(inFiles):
      
     outputFiles = []

     runs = {}

     "Sorting the files"

     for f in inFiles:
        attrs = _metaAttributes.get(f)

        from MetadataHelper import getFileHashValue
        key = getFileHashValue(attrs)

        if not runs.has_key(key):
          runs[key] = [ (attrs.getFilePartition(), f) ]
        else:
          runs[key].append( (attrs.getFilePartition() ,f) )

     for (key, run ) in runs.items():
       run.sort()
       for ( partNum , f ) in run:
         outputFiles.append(f)

     #print "resultant Ordering is ", outputFiles
     return outputFiles



def min(a,b):
   if a > b:
      return b 
   return a

def max(a,b):
   if a > b:
      return a
   return b

class InputAccumulator:

      def __init__(self,maxMergedFileSize):
        self._maxParentFileSize = maxMergedFileSize
        self._currentList = []
        self._currentSize = 0

      def consume(self,inFiles):
        raise CallingAbstractMethodError, "InputAccumulator.consume()"  

      def flush(self):
          l = self._currentList
          s = self._currentSize
          self._currentList = []
          self._currentSize = 0
          return ( l , s )

class StrictSizeByRunInputAccumulator(InputAccumulator):

      def __init__(self,maxMergedFileSize):
        InputAccumulator.__init__(self,maxMergedFileSize)

      def consume(self,inFiles):
        while len(inFiles):
          inFile = inFiles.pop(0)

          fSize = _metaAttributes.get(inFile).getFileSize().getSizeInBytes()
          if ( fSize >  self._maxParentFileSize ):
             raise Exception("Individual file size "+str(fSize) + " of "+inFile+" is bigger than maxParentFileSize") 
          size = self._currentSize + fSize

          if (    size > self._maxParentFileSize \
               or ( len(self._currentList) > 0   \
                   and    _metaAttributes.get(self._currentList[0]).get("runNumber") \
                       != _metaAttributes.get(inFile).get("runNumber") ) ):
             inFiles.insert(0,inFile)
             return self.flush()
          else:
            self._currentList.append(inFile)
            self._currentSize  = size

            if (size == self._maxParentFileSize ):
               return self.flush()
        return ( None , self._currentSize )



class SimpleInputAccumulator(InputAccumulator):
      #Accumulates whatever is passed to it, without any checks i.e. the checks are done externally
      def __init__(self,maxMergedFileSize):
        InputAccumulator.__init__(self,maxMergedFileSize)

      def consume(self,inFiles):
        while len(inFiles):
          inFile = inFiles.pop(0)
          fSize = _metaAttributes.get(inFile).getFileSize().getSizeInBytes()
          self._currentList.append(inFile)
          self._currentSize = self._currentSize + fSize
        return self.flush()



def getNumEvents(infile):
      return _metaAttributes.get(infile).getEventCount()



def getTotalNumEvents(filelist):

      totalEvents = 0
      for f in filelist:
         print "Getting metadata of file %s" %f
         numEvents = getNumEvents(f)
         print "num events = %d" % numEvents
         totalEvents = totalEvents + numEvents
      # end for

      if totalEvents == 0:
         raise Exception("There are no events in the dataset")

      return totalEvents

    
def generateInitialOutputMD(inFiles):

      fileName = inFiles[0]
      attrsStart = _metaAttributes.get(fileName)

      version = attrsStart.get('appVersion')

      runNum = -1
      runDescriptorList = attrsStart.get(SAM.attrRunDescriptorList)
      if runDescriptorList is not None:
          # take the first one:
          runDescriptor = runDescriptorList[0]
          if not runDescriptor.isRunId():
              runNum = runDescriptor.getRunStruct().getRunNumber()
      if ( runNum == -1 ):
        raise Exception("Run number could not be retreived for "+fileName)

      dataStream = attrsStart.get('datastream')
      if ( dataStream == None ):
        dataStream = "all"

      dataTier = attrsStart.get('datatier')
      if ( dataTier == None ):
        raise Exception("Unable to retrieve dataTier from "+fileName)

      print runDescriptor, runNum
      global availableTierPrefixes
  
      try:
        template = "%s_%s_%010d_mrg_%%03d-%%03d.raw_%s" % (availableTierPrefixes[str(dataTier)], dataStream,int(runNum), version ) 
      except KeyError,ex:
         raise Exception("dataTier "+str(dataTier) +" is not found in the naming convention map")

      print "File name template: %s" % template

      attrsEnd = _metaAttributes.get(inFiles[len(inFiles)-1])

      startPart = str(attrsStart.getFilePartition())
      endPart =  str(attrsEnd.getFilePartition())

      print "Start ",startPart,"End",endPart
      outFile = template % ( int(startPart) , int(endPart))
      print "Generated "+outFile

      # find loom block min max range
      lumMin = 0x7fffffff
      lumMax = -1
      for inFile in inFiles:
        inFileAttrs = _metaAttributes.get(inFile)
        tlumMin = inFileAttrs['lumBlockRangeList'][0].getLumMin()
        tlumMax = inFileAttrs['lumBlockRangeList'][0].getLumMax()

        if ( lumMin > tlumMin ):
           lumMin = tlumMin

        if ( lumMax < tlumMax ):
           lumMax = tlumMax

      result = {}
      result['parents'] = inFiles
      result['firstEvent'] = attrsStart.getFirstEvent()
      result['lastEvent'] = attrsEnd.getLastEvent()
      result['totalevents'] = getTotalNumEvents(inFiles)
      result['filename'] = outFile
      result['lumMin'] = lumMin
      result['lumMax'] = lumMax

      return result


def hasChildren(filename):
       _metaAttributes.invalidate(filename)
       fileinfo = _metaAttributes.get(filename)
       parentlist = fileinfo['parents']
       version = fileinfo.get('appVersion')
       tier = fileinfo.get('datatier')
       if fileinfo.getChildren():
           for chld in fileinfo.getChildren(): 
               # check to make sure we are not parent already for the merged children
               chldinfo = _metaAttributes.get(chld)
               if ( chldinfo.get('appVersion') != version ):
                   continue
               return 1
       return 0


def locationsDoNotExist(filename):
   locs = _metaAttributes.getLocations(str(filename))
   if ( len(locs) ):
     return 0
   return 1


def _doesFilterApply(files,filter):
   if ( filter == None ):
      return 0
   else:
      if ( not filter(files[0])):
         return 0
   return 1

def do_isMergedFileAlreadyInSam(filename,filter = None ,strict = 0):
        _metaAttributes.invalidate(filename)
        fileinfo = _metaAttributes.get(filename)
        parentlist = fileinfo.get('parents')
        version = fileinfo.get('appVersion')
        dataTier = fileinfo.get("datatier")

        if ( areVersionsEqual(fileinfo.getChildren(),version,strict) > 0 ): 
            # check to make sure we are not parent already for the merged children
           if ( not _doesFilterApply(fileinfo.getChildren(),filter) ):
              return 1

        #print "Version of ",filename,version
        #print "Parents of ",filename,parentlist
	if parentlist:
            if(len(parentlist)<1):
                raise Exception(filename+" has no parents")

            for parent in parentlist:
                if(len(parent)<1):
                    raise Exception(filename+" has no parents")
                _metaAttributes.invalidate(parent)
                samfileinfo =  _metaAttributes.get(parent)
                siblings = samfileinfo.getChildren()
                #print "Children of ",parent,siblings
                if (siblings):
		    for sib in siblings:
                        #print "Version of ",sib,sibinfo.version,sibinfo.dataTier
                        if (not isVersionEqual(sib,version,strict) ):
                            continue
                        if (sib == filename): 
                            #the same file, possible in the new style parantage, 
			    #not possible in the old one. 
                            #do not merge any of the reco duplicates
                            continue    

                        sibinfo =  _metaAttributes.get(sib)
 
                        if (sibinfo.get('dataTier').find("thumbnail") == -1 or sibinfo.get('dataTier') != dataTier ):
                            continue
                        if (areVersionsEqual(sibinfo.getChildren(),version,strict)== 0 \
                         and fileinfo.get('dataTier').find("unmerged") != -1 ):
                            continue #continue looking if no childdren and new style
                        print sib,sibinfo.getChildren()

                        if (areVersionsEqual(sibinfo.getChildren(), version,strict)==0 \
                        or (not _doesFilterApply(sibinfo.getChildren(),filter) )):
                            return 1
        return 0




def isMergedFileAlreadyInSam(filename,filter = None ):
   res = do_isMergedFileAlreadyInSam(filename,filter,0)
   if res:
     return do_isMergedFileAlreadyInSam(filename,filter,1)
   return res


versionRe = re.compile(""".*p[0-9][0-9]\.[0-9][0-9]\.[0-9][0-9].*""")
versionReStrict = re.compile(""".*_p[0-9][0-9]\.[0-9][0-9]\.[0-9][0-9]_.*""")


def isVersionEqual(filename,version,strict = 0):
   
   if strict:
     re = versionReStrict
   else:
     re = versionRe

   if ( re.match(filename,1) ):
      if ( filename.find(version) != -1 ):
         return 1
      return 0

   fileinfo = _metaAttributes.get(filename)

   if ( fileinfo.get('appVersion') == version ):
       return 1
   return 0
 


def areVersionsEqual(filenames,version,strict = 0):
   if filenames:
       if strict:
           for f in filenames:
               if isVersionEqual(f,version,strict):
                   return 1
       else:
           if len(filenames) > 0:
               return 1
   return 0



def checkDuplicatesInDs(fileNames):

   fileSet = {}
   for f in  fileNames:
      if ( fileSet.has_key(f) ):
         raise Exception("Duplicate file name "+f)
      fileSet[f] = 1

   for filename in fileNames:
      fileinfo = _metaAttributes.get(filename) 
      parentlist = fileinfo.getParents()
      if (parentlist):
          for parent in parentlist:
              _metaAttributes.invalidate(parent)
              samfileinfo =  _metaAttributes.get(parent)
              siblings = samfileinfo.getChildren()
	      if (siblings):
                  for sib in siblings:
                      if ( sib != filename and fileSet.has_key(sib) and samfileinfo.get('datatier') == fileinfo.get('datatier') ):
                          raise Exception("Duplicate file content "+filename+","+sib)

   print "returning from checkDuplicateInDs"

def tests():
  print isMergedFileAlreadyInSam('recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10',locationsDoNotExist)
  print isMergedFileAlreadyInSam("recoT_all_0000176852_009.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42")

  checkDuplicatesInDs(['recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10','recoT_all_0000176852_007.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_004.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_006.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_003.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_001.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_008.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_009.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_005.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_002.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103307563_SamGrid_NumEv-10','recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103582229_SamGrid_NumEv-10'])


  print sortInputFiles(['all_0000176852_009.raw','all_0000194526_095.raw', 'all_0000194526_094.raw', 'recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06', 'all_0000194526_093.raw', 'all_0000194526_092.raw' ,'recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06'])
  print sortInputFiles(['recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10','recoT_all_0000176852_007.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42'])

  print sortInputFiles(['all_0000152186_004.raw','recoT_all_0000152186_003.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC','recoT_all_0000152186_002.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC','recoT_all_0000152186_001.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC'])

  print generateInitialOutputMD(['recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06','recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10','recoT_all_0000176852_007.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42'])
  print generateInitialOutputMD(['recoT_all_0000152186_001.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC','recoT_all_0000152186_002.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC','recoT_all_0000152186_003.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC','all_0000152186_004.raw'])

  print checkDuplicatesInDs(['recoT_all_0000196489_001.raw_p17.05.01_parag_sammy.fnal.gov_101256_576_SamGrid-Testbed_NumEv-250','zeroB_all_0000196489_001.raw_p17.05.01_parag_sammy.fnal.gov_101256_576_SamGrid-Testbed_NumEv-250','recoT_all_0000196489_002.raw_p17.05.01_parag_sammy.fnal.gov_101256_576_SamGrid-Testbed_NumEv-250']) # no duplicates here


  print isMergedFileAlreadyInSam("recoT_all_0000191882_070.raw_p17.03.03_d0farm_fnd0101.fnal.gov_105004_26519_FNAL") # 0
  print isMergedFileAlreadyInSam("recoT_all_0000152186_001.raw_p17.03.03_jenkins_grid.dpcc.uta.edu_002229_20788_UTA-DPCC") # 1
  print isMergedFileAlreadyInSam("zeroB_all_0000196489_005.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250") # 0
  print isMergedFileAlreadyInSam("zeroB_all_0000196489_001.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250") # 0
  print isMergedFileAlreadyInSam("zeroB_all_0000196489_002.raw_parag_sammy.fnal.gov_215835_6639_SamGrid-Testbed_NumEv-250") # 0
  print isMergedFileAlreadyInSam("zeroB_all_0000196489_006.raw_parag_sammy.fnal.gov_215835_6639_SamGrid-Testbed_NumEv-250") # 0
