#!/usr/bin/env python
################################################################################
# Project   : SAM
# Package : 
# $Id: MergeImplementation.py,v 1.49 2007/04/18 15:07:34 parag Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
#
# This file is used to run a project over a setup file configured
# by the user.  Please see project_setup.py for more information.
#
#
################################################################################

import os
import sys
import time
import random
import traceback
import string
import commands
import re

from samProject import Printers
from Sam import sam
import SAMDbServer

import Fetcher
import MergeFlow
import MetadataHelper

from SamStruct.NameOrIdList import NameOrIdList
from SamStruct.SamSize import SamSize
from SamStruct.SamTime import SamTime

DEBUG=0

class MergeImplementation :
   def __init__ (self,argDict):
     try:
	print "just about to be doinit"
        self.doInit(argDict)
        if not DEBUG:
          self.prestageAndMerge()
     except Exception,ex:
        traceback.print_exc(file=sys.stderr)
        self.updateProgress("Top level exception,see err output for details: "+str(ex))
        raise ex

   def doInit(self,argDict):
      # the only global SamMDCache object
      self._print = Printers("merge")      
      # till we get findFileByEvent sorted out in DhGet (make it like in dhGetLite)
      self._withproject = argDict.get('withproject')
      if not self._withproject == None:
         raise Exception("It is not safe to merge files using --with-project\
 option because of some unimplemented functionalities")
 
      self._commonargs = {
                            'withproject' : argDict.get('withproject'),
		            'timeout'     : argDict.get('timeout'),
		            'station'     : argDict.get('station'),
                            'group'       : argDict.get('group'),
			 }
      
      self._jobFileDsId   = argDict.get('jobFilesDatasetId')
      self._jobFileDsName = argDict.get('jobFilesDatasetName')
      self._mergeDsId     = argDict.get('mergeDatasetId')
      self._mergeDsName   = argDict.get('mergeDatasetName')
      self._mergeDim      = argDict.get('dim')
      self._mergeFileList = argDict.get('filename')
      self._maxMergedFileSize = argDict.get('maxMergedFileSize')
      self._appversion = argDict.get('appversion')
      self._jobNumber     = argDict.get('jobNumber')  
      self._facilityname  = argDict.get('facilityname')
      self._skipDuplicatesCheck = argDict.get('skipDuplicatesCheck')
      self._runRecocert = "false"
      self._recocertDurableLocation = ""
      if argDict.has_key('debugLevel'):
          self._debug = int(argDict.get('debugLevel'))
      else:
          self._debug = 0

      
      if (argDict.has_key('runRecocert')):
          self._runRecocert = string.lower(argDict.get('runRecocert'))
          if (self._runRecocert == "true"):
              self._recocertDurableLocation = argDict.get('recocertDurableLocation')
              
      if ( not self._maxMergedFileSize ):
         self._maxMergedFileSize = 1500000000
      else:
          self._maxMergedFileSize = int(self._maxMergedFileSize)
 
      if ( self._jobFileDsId ):
         self._jobFileDsId = int(self._jobFileDsId)

      if ( self._mergeDsId ):
          self._mergeDsId = int(self._mergeDsId)

      if ( self._jobNumber ):
          self._jobNumber = int(self._jobNumber) 

      import jim_sam_storage_negotiator
      self._incomingNode = jim_sam_storage_negotiator.   \
				jim_sam_incoming_storage_negotiator(self._commonargs['station'],"dzero_reco_merge")
      self._outgoingNode = jim_sam_storage_negotiator.	 \
				jim_sam_outgoing_storage_negotiator(self._commonargs['station'],"fss_stager")

      self._commonargs['nodeName'] = self._incomingNode[0]

   def prestageAndMerge(self):
      # stage files to be merged
      cwd = os.getcwd()
      inDir = os.path.join(cwd, "unmerged_files")

      # Create Output File List
      outputDirPrefix = "%s/%s" % (cwd, os.environ['JIM_GLOBAL_JID'])
      working_dirname = "%s-1" % (os.environ['JIM_GLOBAL_JID'])
      outputDir = "%s/%s/D0Copydoom-%s-%s" % (cwd, \
                                              working_dirname, \
                                              self._appversion, \
                                              working_dirname)

      shipOutput = """cd %s; ls -laR %s* > ./created_files.txt; tar cvfz $OUTPUT_FILE `find . -name *log` ./created_files.txt `find . -name *core*` `find . -name events.copy*` `find . -name *macro*` `find . -name import_*py`""" % (cwd, cwd)

      try:      

        numFilesToRelease = 0
        inputFileStream = [] 
        fileIds = Fetcher.getFileIdList(self._mergeDsId,self._mergeDsName)

        fileNames = MetadataHelper.resolveIdsToNames(fileIds)
        
        try:
          MergeFlow.checkDuplicatesInDs(fileNames)
        except Exception,ex:
           traceback.print_exc(file=sys.stderr)
           self.updateProgress("checkDuplicatesInDs failed: "+str(ex))
           raise ex

        sortedFileIds = self.filterMerged(MetadataHelper.resolveNamesToIds(MergeFlow.sortInputFiles(fileNames)))
        samDataStream = Fetcher.FetchStream(\
                            inDir, \
                            sortedFileIds,\
                            self._commonargs)
        toRelease = []
        acc = MergeFlow.SimpleInputAccumulator(self._maxMergedFileSize)

        # stage and setup job files
        if self._jobFileDsId or self._jobFileDsName :
           self.stageJobFiles()

        counter = 0
        while 1:

            if ( self._incomingNode[1] ):
                os.environ["SAM_FCP_QUEUE_NAME"] = self._incomingNode[1]

            dataChunk = None
            if ( len(inputFileStream) == 0 ):
               ( numFilesToFetch, actualFileSetSize ) = \
                   samDataStream.getNumFilesForSize(self._maxMergedFileSize)
               dataChunk = samDataStream.fetch(numFilesToFetch)
               if ( dataChunk == None ):
                   self._print.info("dataChunk is none, last iteration of loop")
                   # merge the tail of the dataset
                   ( filesToMerge, size ) = acc.flush()
                   if ( filesToMerge ):
                       self.merge(filesToMerge,inDir,outputDir,outputDirPrefix)
                       for d in toRelease:
                           d.release()
                   break
               else:
                   self._print.info("prefetched %d files for an est. \
                                    size of %d Bytes" % \
                                    (numFilesToFetch,actualFileSetSize))
                   newData = dataChunk.getData()
                   inputFileStream = inputFileStream + newData
                   toRelease.append(dataChunk)

            self._print.info("Got InputfileStream: "+str(inputFileStream))
            ( filesToMerge , size ) = acc.consume(inputFileStream)
            self._print.info("Files ready to be merged: "+str(filesToMerge)+ \
                             ". Current size "+str(size)+\
                             ". Need to reach size " + \
                             str(self._maxMergedFileSize)+ \
                             " before merging." )


            if ( filesToMerge ):
               numFilesToRelease = numFilesToRelease + len(filesToMerge)
               self.merge(filesToMerge,inDir,outputDir,outputDirPrefix) 







               while len(toRelease):
                  relCandidate = toRelease.pop(0)
                  c = len(relCandidate.getData()) 
                  if ( c > numFilesToRelease ):
                      toRelease.insert(0,relCandidate)
                      break
                  else:
                      relCandidate.release()
                      numFilesToRelease = numFilesToRelease - c
               #self.cleanup(filesToMerge) 
               #site cleanup script will take care of merged files

      except:
          traceback.print_exc(file=sys.stderr)
          os.system(shipOutput)
          raise
          
      os.system(shipOutput)


   def merge(self,filesToMerge,inDir,outputDir,outputDirPrefix):

       #self._print.info("Start merge function")

       unmergedFiles = "unmerged_files.txt" 
       mergedFiles = "merged_files.txt"

       partialFileList = self.createInputFileList(inDir, filesToMerge , unmergedFiles)

       self.updateProgress("Finished staging input file(s): " + \
                           string.join(partialFileList,",") + \
                           " launching merge")

       initialMD = MergeFlow.generateInitialOutputMD ( 
                              MergeFlow.sortInputFiles(partialFileList))

       self._print.info("Initial metadata generated")
       self._print.info(initialMD)
       self._print.info("===================================")

       if ( len(initialMD['parents']) == 0 ):
           raise Exception("Parent list for "+f+" is empty")
   
       self._print.info("Input files are "+str(partialFileList))
       self.createOutputFileList(outputDir,[initialMD['filename']], mergedFiles )

       macroFile = self.createMacro(self._appversion, unmergedFiles, \
                                    mergedFiles, outputDirPrefix, \
                                    partialFileList, [initialMD['filename']] )

       if not DEBUG:
         self.updateProgress("Finished setup, launching d0runjob for merging")
         self.d0runjob(macroFile)
         self.consistencyCheck(outputDir,initialMD)
       else:
         os.system("mkdir -p %s; dd if=/dev/zero of=%s count=%d" % ( outputDir , outputDir+"/"+initialMD['filename'] , 10 ) )

       outFiles = os.listdir(outputDir)

       self._print.info("Discovered following files in directory %s: %s" \
                     %(outputDir,str(outFiles)))


       mdFileName = self.generateMD(outputDir, initialMD , MetadataHelper._samMDCache.get(initialMD['parents'][0]))

       self.storeOne(initialMD,outputDir,mdFileName )

       if (self._runRecocert == "true"):
           try:
               recocert_cmd = "recocert.py --station=%s --group=%s --appversion=%s --timeout=%s --jobFilesDatasetName=%s --mergedFiles=%s/%s --jobNumber=%s --debugLevel=%s --durableLocation=%s" % (self._commonargs['station'], self._commonargs['group'], self._appversion, self._commonargs['timeout'], self._jobFileDsName, outputDir, initialMD['filename'], self._jobNumber, self._debug, self._recocertDurableLocation)

               if ( self._facilityname ):
                 recocert_cmd += " --facilityname="+self._facilityname

               self._print.info("Executing Recocert: %s" % recocert_cmd)
               self.updateProgress("Finished merging, launching recocert")
               rc = os.system(recocert_cmd)
               rc = rc/256
               self.updateProgress("recocert exited with exit code %s" % rc)
               if rc != 0:
                   self._print.warning("Recocert failed. Check logs for details")
           except Exception,ex:
               self._print.warning("Recocert failed. Check logs for details")
               self.updateProgress("recocert exited with non zero exit code")
               traceback.print_exc(file=sys.stderr)

       os.system("""cmd="rm %s";echo Doing,$cmd; $cmd""" % outputDir+"/"+initialMD['filename'])


   def consistencyCheck(self,outputDir,initialMD):

       import commands
       (status , eventCopyFiles ) =  \
		commands.getstatusoutput("find . -name events.copy -type f -print 2>/dev/null")
       if ( status != 0 ):
           raise Exception("Failed to locate event.copy verification file")

       eventCopyFilesArray = eventCopyFiles.split("\n")
       if ( len(eventCopyFilesArray) > 1 and eventCopyFilesArray[1] != "" ):
          raise Exception("Duplicate copies of events.copy file found:"+eventCopyFiles)

       eventCopyFile = eventCopyFilesArray[0]

       if ( eventCopyFile == "" ):
          raise Exception("Consistency check impossible, event copy file is not found")

       self._print.info("Found event copy file :"+eventCopyFile)

       totalevents = initialMD['totalevents']
 
       analyzeEventsCountFile(totalevents,eventCopyFile)
       self._print.info("Checking the contents of %s" % outputDir)
       commands.getstatusoutput("""ls -latr %s*""" % outputDir)
       ( status , corefiles ) = \
		commands.getstatusoutput("""find %s* -name core.* -type f -print 2>/dev/null | grep -v py""" % outputDir)

       if ( status != 256 ):
          raise Exception("Found core files. aborting merge:"+corefiles)

       if ( corefiles != "" ):
          raise Exception("Found core files. aborting merge:"+corefiles)

       os.system("mv %s %s.%s" % ( eventCopyFile , eventCopyFile , initialMD['filename'] ))
 


   def filterMerged(self,fileIds):
      if ( self._skipDuplicatesCheck ):
         return fileIds

      result = []
      for id in fileIds:
        fileName = MetadataHelper._metaAttributes.getFileName(id)
        if ( not MergeFlow.isMergedFileAlreadyInSam(fileName) ):
           result.append(id)
        else:
           self._print.info(fileName+ " is already merged,skipping")
           self.updateProgress("Skipped "+fileName)

      return result



   def checkIfGroupIsMerged(self,fileNames):
      if ( self._skipDuplicatesCheck ):
         return 0

      for fileName in fileNames:
         if ( MergeFlow.isMergedFileAlreadyInSam(fileName) ):
            return 1

      return 0
  


   def untar(self,pattern):
      cmd = "tar xzf %s && rm -f %s" % (pattern,pattern)
      status,stdouterr = commands.getstatusoutput(cmd)
      if status != 0:
	 raise Exception("untar failed %s" % stdouterr)
   ## ==== end of untar


   def getD0runjobDir(self, path):
      import os
      list = os.listdir(path)
      pattern = "d0runjob"
      
      for i in list:
         if i.startswith(pattern):
            return i
      
      raise Exception("Unable to get the D0_RUNJOB_DIR")
   ## ==== end of getD0runjobDir


   def stageJobFiles(self):
      try :

         if ( self._incomingNode[1] ):
             os.environ["SAM_FCP_QUEUE_NAME"] = self._incomingNode[1]

         cwd = os.getcwd()

         stream = Fetcher.FetchStream(\
                      cwd,\
                      Fetcher.getFileIdList(self._jobFileDsId, \
                                            self._jobFileDsName),\
                      self._commonargs)
         while 1:
            chunk = stream.fetch(5)
            if ( chunk == None ):
               break

         self.updateProgress("Finished staging job files")
         self._print.info("Setting up d0 code")
         self.untar("d0_*.tar.gz")
         os.environ['SRT_DIST'] = "%s/RTE_runjob" % cwd
         # remove bad libs
         self._print.info("Setting up d0runjob")
         self.untar("d0runjob*.tar.gz")
         self.updateProgress("Finished unpacking d0runjob and d0_RTE")
         os.environ['D0_RUNJOB_DIR'] = cwd + "/" + self.getD0runjobDir(cwd)
      except Exception,ex :
         self._print.error("Failed to stage job files")
       	 raise ex
   ## ==== end of setupJobFiles


   # This function will take as input either a list of files,dataset name
   # or a dim query. It will get all the files for the input from SAM
   # and returns the list of files staged
   # This function creates a text file containig all the input files
   # to be fed to d0_runjob
   def createInputFileList(self,inDir, filelist, filename):
      if len(filename) == 0:
         raise Exception("Invalid filename specified for input file")

      actualFiles = os.listdir(inDir)
      actualFiles.sort()
      if not len(actualFiles) :
         raise Exception("No files under %s" % inDir)

      actualFilesMap = {}
      for actualFile in actualFiles:
          actualFilesMap[actualFile] = 1

      finalList = []
      for f in filelist:
         if ( actualFilesMap.has_key(f) ):
            finalList.append(f)
         else:
	    raise Exception("File :"+f+" is missing on worker node, while reported delivered")

      if not len(finalList) :
         raise Exception("No new files under %s" % inDir)
   
      try :
         file = open(filename, 'w')
         for f in finalList :
            file.write("%s/%s\n" %(inDir, f) )
         file.close()
      except Exception,ex:
         self._print.error("Failed to write InputFileList")    
         raise ex
	 
      return finalList
   # ==== end of createInputFileList


   # This function creates a text file that contains the names
   # of the output files to be produced by d0_runjob
   def createOutputFileList(self,outputDir, outFileList , filename) :

      cleanupCmd = "rm -rf %s" % outputDir
      self._print.info("Executing %s" % cleanupCmd)
      os.system(cleanupCmd)
   
      try :   
	 f = open(filename, "w")
         for i in outFileList: f.writelines("%s/%s" %(outputDir,i))
         f.close()
      except Exception,ex :
         self._print.error("Failed to create output file list")
	 raise ex
	 
      self._print.info("Generated file list: %s" % str(outFileList) )
      sys.stdout.flush()
   ## ==== end of createOutputFileList


   # This function creates a macro for running a merging job
   def createMacro(self, d0release, unmergedFiles, mergedFiles, destDir, unmergedFilesList, mergedFilesList):
      working_dir = os.getcwd()
      srt_dist = os.environ['SRT_DIST']
      srt_dist_dir = srt_dist + "/releases/" + d0release
      for i in range(0, len(unmergedFilesList)):
          try:
              print "-> FINDING: %s in %s " % (unmergedFilesList[i],working_dir)
              os.system("find . -name %s" % (unmergedFilesList[i]))
          except:
              print "======= ERROR FINDING %s" % (unmergedFilesList[i])
              pass
          unmergedFilesList[i] = "%s/unmerged_files/%s" % (working_dir, unmergedFilesList[i])
      unmergedFiles_list = str(unmergedFilesList)
      #mergedFiles_list = mergedFilesList[0]
      mergedFiles_list = str(mergedFilesList)
      #mergedFiles_list = mergedFiles
      gjid = os.environ['JIM_GLOBAL_JID']

      print "-------------------> Additional Info"
      print "%s %s %s %s %s %s" % (working_dir, srt_dist, srt_dist_dir, unmergedFiles_list, mergedFiles_list, gjid)

      macro="""
import os

locald0 = os.path.expandvars("%s")

from d0rjcore.d0userinterface import createUserInterface
d0rj = createUserInterface(
    JobName = "%s",
    JobCacheArea = "%s",
    OptionalMonitors = ['jim'],
)

d0copydoom = d0rj.attachExecutable("D0Copydoom")
d0copydoom["InputFiles"] = eval("%s")
d0copydoom["OutputFiles"] = eval("%s")
d0copydoom["ProjectVersion"] = "%s"
d0copydoom["DistroDirectory"] = "%s"
d0copydoom["CommandLineOpts"] = "-status_file events.copy -status_level 1"

# Register general job manager with BashJobOperator in order to produce
# the required job and task scipts
bashjob = d0rj.attachConfigurator("BashJobOperator")
bashjob.registerManager(d0rj.getManager())

prov = d0rj.attachConfigurator("ProvenanceProcessor")
prov.registerConfigurator(d0copydoom)

run = d0rj.attachConfigurator("RunInteractive")
run.registerManager(d0rj.getManager())

for i in range(0, 1):
  d0rj.run()
""" % (srt_dist_dir, gjid, working_dir, unmergedFiles_list, mergedFiles_list, d0release, srt_dist)

      # The name of the macro should not contain absolute path
      # mc_runjob doesnt like absolute paths

      macroFile ="merge.macro"
      try :   
	 f = open(macroFile, 'w')
         f.write(macro)
         f.close()
      except Exception, ex :
         self._print.error("Failed to write macro file")
         raise ex       

      print "Macro dumped in file %s" % macroFile
      print "------------------------------------"
      print "         Generated Macro            "
      print "------------------------------------"
      print "%s" % macro
      print "------------------------------------"
      sys.stdout.flush()

      return macroFile   
      
   ## ==== end of createMacro      


   # call d0runjob
   def d0runjob(self,macroFile):
      # Setup the environment for d0runjob
      os.environ['PYTHONPATH'] = os.environ['PYTHONPATH'] + ":" + os.environ['D0_RUNJOB_DIR'] + "/Python"
      os.environ['SHAHKAR_FACTORY_EXT'] = os.environ['D0_RUNJOB_DIR'] + "/xml/d0factory.xml"
      
      try :
         if not os.environ['D0_RUNJOB_DIR']:
            raise Exception("Need D0_RUNJOB_DIR env variable set or specify jobfiles dataset")

         print "Invoking d0runjob to merge files"
         cmd = "python %s" %(macroFile)
         sys.stdout.flush()
         print "==============================================\n"
         print os.environ
         print "==============================================\n"
         print "Executing %s" % (cmd)
         status = os.system(cmd)
         print "Exit code = %d" % status
         print "\n=============================================="

         if status != 0:
            raise Exception("d0runjob returned an invalid status code")
      
      except Exception,ex:
         self._print.error("d0runjob failed !")
	 raise ex	    
   ## ==== end of d0runjob


   def makeSamDbTime(self,secs):
      return time.strftime("%m/%d/%Y %H:%M:%S", time.gmtime(secs))
   ## ==== end of makeSamDbTime


   def generateMD(self,outDir, basicMD, samMD):

      fileName = basicMD['filename']
      self._print.info("SAM Template MD :"+str(samMD))
      fileStruct = None
      try:
          fileStruct = os.stat(os.path.join(outDir, fileName))
      except OSError,ex:
          self.updateProgress("job generated no output file "+fileName)
          raise ex

      fileSize   = fileStruct[6]/1024
      fileCreateTime = fileStruct[9]
      fileModifiedTime = fileStruct[8]

      header = """from SamFile.SamDataFile import DerivedDetectorFile
from SamFile.SamDataFile import ImportedSimulatedFile
from SamStruct.SamSize import SamSize
from SamStruct.CRC import CRC
from SamStruct.NameOrIdList import NameOrIdList
from SamUtility.CaseInsensitiveDictionary import CaseInsensitiveDictionary
from SamStruct.SamTime import SamTime
from SamStruct.ApplicationFamily import ApplicationFamily
from SamStruct.Params import Params
from SamStruct.RunDescriptorList import RunDescriptorList
from SamStruct.RunDescriptor import RunDescriptor
from SamStruct.ParamValue import ParamValue
from SamStruct.DataType import DataType
from SamStruct.LumBlockRangeList import LumBlockRangeList
from SamStruct.LumBlockRange import LumBlockRange
from SamStruct.SamList import SamList
import SAM
"""
      newMDStr= header + "newMD = "+str(samMD)

      exec newMDStr

      newMD["fileName"] = fileName
      newMD["parents"] = NameOrIdList(basicMD["parents"])
      newMD[SAM.attrFirstEvent] = basicMD[SAM.attrFirstEvent]
      newMD[SAM.attrLastEvent] = basicMD[SAM.attrLastEvent]
      newMD["eventCount"] = basicMD['totalevents']
      newMD["fileSize"] = SamSize(str(fileSize)+"K")
      newMD["startTime"] = SamTime(fileCreateTime)
      newMD["endTime"] = SamTime(fileModifiedTime)
      newMD["lumBlockRangeList"] = LumBlockRangeList([LumBlockRange(basicMD['lumMin'],basicMD['lumMax'])])
      newMD['sourceSplit'] = 0
      newMD['destMerge'] = 0
 
      if ( self._facilityname ):
          if ( newMD.get('params') ):
              if ( newMD.get('params').get('Global') ):
                  newMD.get('params').get('Global')['facilityname'] = self._facilityname
              else:
                  newMD.get('params')['Global'] = { 'facilityname' : self._facilityname } 
          else:
              newMD['params'] = Params({'Global' : {'facilityname': self._facilityname } } ) 
  
      if (newMD.get('params')):
          if (newMD.get('params').has_key('global')):
              if (newMD.get('params').get('global').has_key('runjobversion')):
                  del(newMD.get('params').get('global')['runjobversion'])

      dataTierMergeMap = { 'unmerged-filtered-raw' : 'filtered-raw' , 'unmerged-thumbnail' : 'thumbnail' }
      newMD["dataTier"] = dataTierMergeMap[samMD.get('datatier')]
     
      newMD["crc"] = sam.calculateFileCrc(path=os.path.join(outDir, fileName))
      self._print.info("SAM Generated MD :"+str(newMD))

      mdFileName = "%s/import_%s.py" % (outDir, fileName)
      print "-----------\nGenerated Metadata is in \n----------"
      print mdFileName
      try :
         mdFile = open(mdFileName, 'w')
         mdFile.write(header + "t=" + str(newMD))
         mdFile.close()
      except Exception, ex :
         self._print.error("Failed to write metadata file")
         raise ex 

      print "-----------\nGenerated Metadata file reads \n----------"
      try:
         f = open(mdFileName, 'r')
         for line in f.readlines():
             print line
         f.close()
      except:
          pass
      return mdFileName

############### OLD CODE

      processedMD = samMD.v
      """samMDCopy = samMD

      set_attribute(samMDCopy,"first",basicMD['beginEvent']) 
      set_attribute(samMDCopy,"last",basicMD['endEvent'])
      set_attribute(samMDCopy,"num",basicMD['totalevents'])
      set_attribute(samMDCopy,"parentFiles",basicMD['parents'])

      set_attribute(samMDCopy,"fileName",basicMD['filename'])
      set_attribute(samMDCopy,"startTime",fileCreateTime)
      set_attribute(samMDCopy,"endTime",fileModifiedTime)

      import SAM 
      samFileSize = SAM.ByteSize(fileSize,SAM.prefix_kilo)
      samFileSize.number = fileSize
      samFileSize.prefix = SAM.prefix_kilo
      set_attribute(samMDCopy,"fileSize",samFileSize)

      sam_crc = SAM.CRC_Data(SAM.CRC_UnknownValue,SAM.CRC_TC_Adler32)
      set_attribute(samMDCopy,"crcData",sam_crc)

      try:
        set_attribute(samMDCopy,"version",self._appversion)
      except Exception,ex:
        self._print.warning(str(ex))
      """ 
	 
      mdText = """from import_classes import *
TheFile=ProcessedFile(
         name= '%s',
         sizeK = %d,
         events = Events(%d,%d,%d),
         stream = 'Unstreamed',
         tier = 'thumbnail',
         start_time ='%s',
         end_time ='%s',
         parents = ['%s'],
         pid = %d) 
""" % ( fileName , fileSize , basicMD['beginEvent'], basicMD['endEvent'] , basicMD['totalevents'] , fileCreateTime , fileModifiedTime , string.join(basicMD['parents'],"""','"""), processedMD.cpid)

      mdFileName = "%s/import_%s.py" % (outDir, fileName)
      try :
         mdFile = open(mdFileName, 'w')
         mdFile.write(mdText)
         mdFile.close()
      except Exception, ex :
         self._print.error("Failed to write metadata file")
         raise ex

      return mdFileName

   ## ==== end of generateMD


   def storeOne(self,basicMD,outDir, mdFileName):
      self._print.info("Trying to store the merged file with metadata in %s " % mdFileName )

      fileName = basicMD['filename']

      if ( self.checkIfGroupIsMerged(basicMD['parents']) ):
         self._print.info("File "+fileName+" can not be merged because its parents" +\
                                           " have been merged already")
         self.updateProgress("Skipped "+fileName)
         return 
 
      if ( self._debug ):
         self._print.info("File "+fileName+" will not be stored in debug mode")
         # Additional info
         print "---------------------------------------------------------------"
         print " FOLLOWING COMMANDS WERE UNEXECUTED AS THIS IS TEST RUN"
         print "---------------------------------------------------------------"
         print "Printing from location %s " % (os.getcwd())
         print "Using outgoing node : %s and fcp q %s " % ( self._outgoingNode[0] , self._outgoingNode[1] )
         print "---------------------------------------------------------------"
         return

      self._print.info("Using outgoing node : %s and fcp q %s " % ( self._outgoingNode[0], self._outgoingNode[1] ) )

      try:
          filepath = os.path.join(outDir,fileName)

          if self._outgoingNode[1]:
              os.environ['SAM_FCP_QUEUE_NAME'] = self._outgoingNode[1]

          if self._outgoingNode[0]:
              transferRequestId = sam.storeFile(sourceFile=filepath, \
                                                waitForCompletion=1,\
                                                temporaryTransferLocation=self._outgoingNode[0],\
                                                retryMaxCount=3,\
                                                descriptionFile=mdFileName)
          else:
              transferRequestId = sam.storeFile(sourceFile=filepath, \
                                                waitForCompletion=1,\
                                                retryMaxCount=3,\
                                                descriptionFile=mdFileName)
      except Exception,ex:
          traceback.print_exc(file=sys.stderr)
          self._print.error("Failed to store file with metadata in %s" % \
                            mdFileName)
          raise "Failed to store file with metadata in %s" % mdFileName

      sys.stderr.flush()
      sys.stdout.flush()

      locList = MetadataHelper.getFileLocation(fileName)
      if ( not len(locList) ):
         raise Exception("File store finished but no location has been registered for "+fileName)

      self.updateStateInDB("stored",fileName)
   ## ==== end of generateMDandStoreOne


   def updateStateInDB(self,state,filename):
      import time
      updateStr = '<output_file name=\"%s\" time=\"%s\" time_epoch=\"%s\" state=\"%s\"/>' % (filename, time.strftime("%b %d %Y %H:%M:%S %Z", time.localtime()), str(int(time.time())), state)
      self.updateDB(updateStr)
   ## ==== end of updateStateInDB



   def updateProgress(self,progressStr):
      updateStr = '<progress created=\"%s\" time_epoch=\"%s\">%s</progress>' % ( time.strftime("%b %d %Y %H:%M:%S %Z", time.localtime()) , str(int(time.time())), progressStr )
      self.updateDB(updateStr)
   ## ==== end of updateProgress


   def updateDB(self,updateStr):
      """Update the JIM XML DB"""
      try:
         gjid = os.environ['JIM_GLOBAL_JID']
         cjid = os.environ['JIM_CLUSTER_JID']
         ljid = os.environ['JIM_LOCAL_JID']
         url  = os.environ['XMLDB_SERVER_URL']
      except Exception, ex:
         traceback.print_exc(file=sys.stderr)
         self._print.warning("%s env variable missing" % ex) 
         self._print.warning("Cannot update XML database")
         sys.stderr.flush()
         return
 
      try:
         self._print.info("Updating XML database")
         jobCollection = "/db/jobs/globalJobs1"
         cmd = """xmldb_cmd id -c %s -n %s_%s -x '/local_job[@Id="%s"]' -url %s -xs '%s'""" % (jobCollection, gjid, ljid, ljid, url, updateStr)
         self._print.info("Executing: %s" % cmd)
         os.system(cmd)
         #from xmldb_client import idmgr
         #jobCollection = "/db/jobs/%s" % (gjid)
         #db = idmgr.idmgr(url, jobCollection)
         #db.UpdateLocal("append", gjid, cjid, ljid, updateStr)
      except Exception, ex:
         traceback.print_exc(file=sys.stderr)
         self._print.warning(ex) 
         self._print.warning("Could not Update XML Database... will continue")
         sys.stderr.flush()
         pass
   ## ==== end of updateDB


   def cleanup(self,tmbList):

      if ( self._debug ):
         self._print.info("Skipping cleanup in debug mode")
         return

      self._print.info("Cleaning up source files from SAM")
      print tmbList  
      if not len(tmbList):
         self._print.warning("No thumbnails to clean")
         sys.stderr.flush()
	 return
      # end if
      
      # tmbs that have been merged
      tmbs2remove = self.getTmbsToRemove(tmbList)
      if not len(tmbs2remove):
         self._print.warning("There are no thumbnails to remove")
         sys.stderr.flush()
         return
         
      self._print.info("Removing files %s" % str(tmbs2remove))
      for tmb in tmbs2remove:
         self.removeTmb(tmb)

      # end for 	 
   ## ==== end of cleanup

   
   def removeTmb(self,tmb):
      try :
         locs = self._sam.call("locate", {'args':tmb})
	 ("Removing locations for file %s" % tmb)
         for loc in locs:
            self._print.info("Removing location: %s" % loc)
            self._sam.call("eraseFileLocation",{'file':tmb,\
	                                        'loc' :loc})
	 # end for 					
      except Exception,ex:
         self._print.warning(ex)
	 self._print.warning("Error while removing locations for file %s" % tmb)
         sys.stderr.flush()
	 pass
   ## ==== end of removeTmb   






   def getTmbsToRemove(self,tmbList):
      result = []
      for tmb in tmbList:
         if MergeFlow.hasChildren(tmb):
	    result.append(tmb)
      # end for	    
      return result
   ## ==== end of getTmbsToRemove


   def test(self):
     # metadata generator test
     from Sam import sam
     #samMD = sam.getMetadata(filename="zeroB_all_0000196489_002.raw_parag_sammy.fnal.gov_215835_6639_SamGrid-Testbed_NumEv-250")
     samMD = sam.getMetadata(filename="recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10")
     import SAM

     import MergeFlow
     #basicMD = MergeFlow.generateInitialOutputMD([ 'zeroB_all_0000196489_002.raw_parag_sammy.fnal.gov_215835_6639_SamGrid-Testbed_NumEv-250' , 'zeroB_all_0000196489_006.raw_parag_sammy.fnal.gov_215836_6639_SamGrid-Testbed_NumEv-250'  ] )
     basicMD = MergeFlow.generateInitialOutputMD([ 'recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10','recoT_all_0000176852_007.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42' ] )
     print basicMD
     outDir = "/tmp"
     import os
     os.system("mkdir -p %s; dd if=/dev/zero of=%s count=%d" % ( outDir , outDir+"/"+basicMD['filename'] , 10 ) )
     mdFileName = self.generateMD(outDir, basicMD, samMD) 
     os.system("cat "+mdFileName)

def set_attribute(obj,attrname,attrvalue):
   if ( not find_and_set_attribute(obj,attrname,attrvalue) ):
      raise Exception("Failed to find and set attr "+attrname+"in "+str(obj))



def find_and_set_attribute(obj,attrname,attrvalue):
   try:
    for ( aname, avalue ) in obj.__dict__.items():
       if ( aname == attrname ):
          if ( type(avalue) == type(attrvalue) ):
            obj.__dict__[attrname] = attrvalue
            return 1
       if ( find_and_set_attribute(avalue,attrname,attrvalue) ):
          return 1
   except:
     return 0



def analyzeEventsCountFile(totalevents,eventCopyFile):

       import commands
       ( status , xmlString ) = commands.getstatusoutput("""sed 's/: /="/g' %s | sed 's/\(..*\)$/\\1"/g' | sed 's/ /_/g' | sed 's/Input/<element Input/g' | sed 's?Status\(..*\)$?Status\\1 />?g' | sed 's/Output/<element Output/g'""" % eventCopyFile )

       if ( status != 0 ):
           raise Exception("Failed to parse events.copy file content")

       from xml_common_lib import PyXmlObject

       try:
          print xmlString
          eventData = PyXmlObject.PyXmlObject("<events>"+xmlString+"</events>")
       except Exception,ex:
          raise Exception("Format of events.count is not as expected "+str(ex))

       if ( len(eventData.events[0].element) < 2 ):
           raise Exception("Format of events.count is not as expected , too little recordrs found, expected at least 2")

       totalInputEvents = 0
       totalOuputEvents = 0
       for record in eventData.events[0].element:
           try:
             fileName = str(record.Input_file[record.Input_file.rfind("/")+1:])
             totalInputEvents = totalInputEvents + int(record.Total_events)

             realEventCount = MetadataHelper._metaAttributes.get(fileName)["eventCount"]
           except AttributeError,ex:
             fileName = record.Output_file[record.Output_file.rfind("/")+1:]
             totalOuputEvents = int(record.Total_events)
             realEventCount = totalevents

           if ( realEventCount != int(record.Total_events)  ):
              raise Exception("Failed to match event numbers. " + record.Total_events +\
                              " expected from metadata of "+fileName)

           if ( record.Status != "closed" ):
              raise Exception("Status of "+fileName + " is not closed")

       if ( totalOuputEvents != totalevents ):
           raise Exception("Output file record with "+str(totalevents)+\
                           " events is not found in events.copy") 

       if ( totalInputEvents != totalOuputEvents ):
           raise Exception("Sum of all input event counts is "+str(totalInputEvents)+\
                           " but ouput file has "+str(totalOuputEvents)+\
                           " ,mismatch, aborting merge")


if __name__ == "__main__":

  analyzeEventsCountFile(21372,"./sample.1") 

  #MergeImplementation({ "station"  : "samgfarm" , "mergeDatasetId" : 193978 , 'maxMergedFileSize' : 818952})

  fileNames = ['recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10','recoT_all_0000176852_007.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_004.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_006.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_003.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_001.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_008.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_009.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_005.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_002.raw_p17.01.00_d0farm_fnd0101.fnal.gov_115959_27712_SamgridFarm_NumEv-42','recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103307563_SamGrid_NumEv-10','recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103582229_SamGrid_NumEv-10']

  #['all_0000176852_009.raw','all_0000194526_095.raw', 'all_0000194526_094.raw', 'all_0000194526_093.raw', 'all_0000194526_092.raw' ]

  print MergeFlow.sortInputFiles(fileNames)

  fileIds = MetadataHelper.resolveNamesToIds(MergeFlow.sortInputFiles(fileNames))
  print fileIds

  print MetadataHelper.resolveIdsToNames(fileIds)


## ============== end of class MergeImplementation ============== ##
