import os
import sys
import time
import random
import traceback
import string
import commands
import re

from Sam import sam
import SAM
import SAMDbServer
from samProject import Printers
from SamFile.SamDataFile import SamDataFile

class SamMDCache:

    def __init__(self):
        self._md = {}
        self._print = Printers("SamMDCache")

    def get(self, filename):
        if self._md.has_key(filename):
            #cache hit
            return self._md[filename]
        try :
	   fileInf = sam.getMetadata(fileName=filename)
           self._md[filename] = fileInf
        except Exception,ex :
           self._print.error("Failed to get metadata for file %s" % filename)
	   print ex
           raise ex

        return self._md[filename]

global _samMDCache
_samMDCache = SamMDCache()

import re
global r
r = re.compile(".*all.*_[0-9]+_([0-9][0-9][0-9])\.raw.*")

def getFilePartition(filename):
    global r
    try:
        groups = r.match(filename)
        part = groups.group(1)
        return int(part)
    except  Exception,ex:
        raise Exception("Filename "+filename+" did not match .*all.*_[0-9]+_([0-9][0-9][0-9])\.raw.* pattern")

def getChildrenListAdHoc(fileid,version = None):
  if ( fileid == None ):
     raise Exception("Expected fileId parameter set  in the metadata")

  from SamUtility.DbDerivedClient import DbDerivedClient
  server = DbDerivedClient('DbStations')

  if ( not version ):
    children_q = "select file_name from data_files,file_lineage,file_content_status where data_files.file_id = file_lineage.file_id_dest and file_lineage.file_id_source=%s and file_content_status.file_content_status = 'good' and file_content_status.file_content_status_id = data_files.file_content_status_id" % str(fileid)
  else:
    children_q = "select file_name from data_files,file_lineage,application_family,file_content_status where data_files.file_id = file_lineage.file_id_dest and application_family.APPL_FAMILY_ID = data_files.APPL_FAMILY_ID and application_family.VERSION = '%s' and file_lineage.file_id_source = %s and file_content_status.file_content_status = 'good' and file_content_status.file_content_status_id = data_files.file_content_status_id" % (str(version), str(fileid))

  children_q_results = server.query(children_q)

  result = []

  for child_r in children_q_results:
    result.append(str(child_r[0]))

  return result  

########## Start - class ExtendedAttributeListConstructor ################
class ExtendedAttributeListConstructor:
      #def __init__(self,**args):
      def __init__(self, filename):
         #filename = args['filename']
         self._o = sam.getMetadata(filename=filename) 

         parentsWrapper = self._o.get("parents")
         if ( parentsWrapper and len(parentsWrapper) ):
             self._o["parents"] = map( lambda x: x.name , parentsWrapper )
             #print parentsWrapper[0]
         else:
             self._o["parents"] = []

         childrensWrapper = self._o.get("children")
         if ( childrensWrapper and len(childrensWrapper) ):
             self._o["children"] = map( lambda x: x.name , childrensWrapper )
         else:
             self._o["children"] = getChildrenListAdHoc(self._o['fileid'])
 
         #print filename,self._o.get('appVersion'),"Children",self._o["children"]

         if ( self._o.getFilePartition() == None ):
         # partion is not set, agreement is to get one from parent
           try:
               if ( self._o.getParents() and len(self._o.getParents()) != 0 ):
                   parents = self._o.getParents()
                   parentMd = sam.getMetadata(filename=parents[0])
                   if ( parentMd.getFilePartition() != None ):
                     self._o.setFilePartition(parentMd.getFilePartition())
                   else:
                     self._o.setFilePartition( getFilePartition(parents[0]) )
               else:
                   self._o.setFilePartition( getFilePartition(filename) )
           except Exception,ex:
               if ( str(ex).find("did not match") == -1 ):
                   raise

         try:
           for ( keyName,keyVal ) in self._o["params"]["global"].items():
             if ( keyName == "facilityname" ):
               self._o['facilityname'] = keyVal
         except KeyError,ex:
           pass


      def get (self):
         return self._o

########### End class ExtendedAttributeListConstructor #######


def getLocations(**args):
       filename = args['filename']
       return sam.locate(arg=[filename])

############### Start of Class MetaAttributes################
class MetaAttributes:

    def __init__(self):
        self._attrCache = {}
        self._fileIdMap = {}
        self._fileNameMap = {}
        self._locsCache = {}

    def invalidate(self,filename):
         if ( self._attrCache.has_key(filename) ):
            del self._attrCache[filename]
         if (  self._locsCache.has_key(filename) ):
            del self._locsCache[filename]


    def get(self,filename): 
         if self._attrCache.has_key(filename):
            return self._attrCache[filename]
	 o = ExtendedAttributeListConstructor(filename).get()
         self._attrCache[filename] = o
         return o


    def getLocations(self,filename):
          if self._locsCache.has_key(filename):
             return self._locsCache[filename]
          locs = sam.locate(args=[filename])
          self._locsCache[filename] = locs 
          return locs

    def getFileName(self,id):
         if self._fileIdMap.has_key(id):
           return self._fileIdMap[id] 
         fileName = self.mapFileIdToName( id )
         self._fileIdMap[id] = fileName
         self._fileNameMap[fileName] = id
         print fileName
         return fileName 


    def getFileId(self,name):
         if self._fileNameMap.has_key(name):
            return self._fileNameMap[name]
         fileId =  self.mapFileNameToId(name)

         self._fileNameMap[name] = fileId
         self._fileIdMap[fileId] = name
         return fileId


    def getByID(self,id):
         return self.get(self.getFileName(id))


    def mapFileIdToName(self, fileId):
          s = sam.getMetadata(fileId=fileId)
	  return s.get("filename")
	 

    def mapFileNameToId(self, fileName):
	  s = sam.getMetadata(filename=fileName)
	  return s.get("fileId")

################## END OF Class MetaAttributes #####################

def resolveIdsToNames(idList):
   result = []
   for i in idList:
      result.append(_metaAttributes.getFileName(i))
   return result



def resolveNamesToIds(nameList):
   result = []
   for i in nameList:
      result.append(_metaAttributes.getFileId(i))
   return result



def getRunStreamKey(attrs):
  runDescriptorList = attrs.get(SAM.attrRunDescriptorList)
  if ( runDescriptorList != None ):
    return str(runDescriptorList[0].getRunStruct().getRunNumber()) + str(attrs.get('datastream')) + str(attrs.get('datatier'))
  else:
    return str(attrs.get('datastream'))

def getFileHashValue(attrs):
   runStremKey = getRunStreamKey(attrs)
   facility = attrs.get('facilityname')
   return runStremKey + str(facility)

def getNumFilesFromListForSize(idList,totSize):
   sizeCounter = 0
   numFilesCounter = 0
   
   #if the list is empty, no file needs to be fetced
   if not idList:
     return (0,0)
     
   currKey = getFileHashValue(_metaAttributes.getByID(idList[0]))

   for i in idList:
     fileSize = _metaAttributes.getByID(i).getFileSize().getSizeInBytes()
     #Are we above the total size if we included this file? 
     tKey = getFileHashValue(_metaAttributes.getByID(i))

     if ( sizeCounter + fileSize ) > totSize or ( tKey != currKey ):
       break
     #The total set is below the totSizeDB: let's count the file in
     numFilesCounter = numFilesCounter + 1
     sizeCounter = sizeCounter + fileSize
     #print "after %d, size is %d" % (numFilesCounter, sizeCounter)
   #We should get at least 1 file to make progress, even if the single file is above totSize
   if numFilesCounter == 0:
     numFilesCounter = 1
     sizeCounter = fileSize

   return (numFilesCounter, sizeCounter)

def getFileLocation(fileName):
    locs = sam.locate(args=[fileName])
    return locs

#print getFileLocation('recoT_all_0000176852_001.raw_p17.01.00_anoopr_test_1103041631_SamgridFarm_NumEv-10')

global _metaAttributes
_metaAttributes = MetaAttributes()

def test():
  from MergeFlow import sortInputFiles
  fileList=['recoT_all_0000194380_084.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_094.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_088.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'zeroB_all_0000196489_001.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250',\
            'recoT_all_0000194380_003.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_010.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_014.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_006.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_016.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_018.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06',
            'recoT_all_0000194380_021.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_027.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_024.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_006.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'zeroB_all_0000196489_002.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250',\
            'recoT_all_0000194380_016.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_018.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_021.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_2_0000215235_mrg_071-077.raw_p17.09.00',\
            'zeroB_all_0000196489_005.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250',\
            'recoT_all_0000194380_027.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_024.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_028.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_029.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'zeroB_all_0000196489_006.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250',\
            'recoT_all_0000194380_034.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_036.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_0000194380_031.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin',\
            'recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06',
           ]
  output = ""
  sortedList = sortInputFiles(fileList)
  while len(sortedList):
    (numFiles, size ) = getNumFilesFromListForSize(resolveNamesToIds(sortedList),1500000000)
    output += str(sortedList[:numFiles])+"  to merge, "+str(size)+"\n"
    sortedList = sortedList[numFiles:]

  ethalon = """['recoT_all_0000194380_034.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_036.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin']  to merge, 429537796.0
['recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06', 'recoT_all_1_0000215235_mrg_099-099.raw_p17.09.06']  to merge, 282241024.0
['recoT_all_0000194380_003.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_006.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_006.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_010.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_014.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_016.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin']  to merge, 1324616231.0
['recoT_all_0000194380_016.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_018.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_018.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_021.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_021.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_024.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin']  to merge, 1306705403.0
['recoT_all_0000194380_024.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_027.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_027.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_028.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_029.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_031.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin']  to merge, 1303487808.0
['recoT_all_0000194380_084.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_088.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin', 'recoT_all_0000194380_094.raw_p17.02.00_anoopr_samadams.fnal.gov_112403_29686_Wisconsin']  to merge, 614329420.0
['recoT_all_2_0000215235_mrg_071-077.raw_p17.09.00']  to merge, 1123821568.0
['zeroB_all_0000196489_001.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250', 'zeroB_all_0000196489_002.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250', 'zeroB_all_0000196489_005.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250', 'zeroB_all_0000196489_006.raw_parag_sammy.fnal.gov_215834_6638_SamGrid-Testbed_NumEv-250']  to merge, 1201706.0
"""
  if ( output != ethalon ):
     print output
     print ethalon
     raise Exception("test failed")

if __name__ == '__main__':
  print _metaAttributes.get("all_0000188479_100.raw").childrenList 
  test()
