#!/usr/bin/env sampy
import string
import cgi
import time
import sys
import os
import signal

TRUE_=1
FALSE_=0

RETRY_MAX_COUNT = 1     # retry ONLY ONE TIME (to recover from rebooted servers)
CORBA_CALL_TIMEOUT = 3  # do not wait forever for results on the web

try:
    docRoot = os.environ["DOCUMENT_ROOT"]
    #special consideration for NUMI, running on an AFS web server,
    # where product setups are not done before-hand.
    #  code contribued by Liz Buckley-Geer
    if string.find(docRoot, 'numi') != -1:
        sys.path.insert(0, '/afs/fnal/ups/ups/v4_7_2/SunOS+5/ups')
        from ups import getUps
        ups = getUps()
        ups.setup('sam')
except KeyError:
    docRoot = ""
 

from SamUtility import FastCgiHandler
from SamException import SamExceptions
from SamCommand import CommandLineArgParser
from SamCommand.SamCommandInterface import ATTR_RETRY_MAX_COUNT, ATTR_CORBA_CALL_TIMEOUT

def getTb(excInfo):
    output  = "Exception: %s %s\n" % (excInfo[0], excInfo[1])
    output += "Traceback:\n"
    import traceback
    tb = traceback.format_tb(excInfo[2])
    for line in tb:
        output += "%s" % line
    return output
#================================================================================

class DumpSam:
    def __init__(self):
        from Sam import sam
        self.sam = sam
        
    def handle(self, form):
        result = "No results were processed."
        try:
            (dumpster, argdict) = self._parseForm(form)
            if dumpster is not None:
                result = self._invokeDumpster(dumpster, argdict)
            else:
                result = "Garbled command, I do not understand what you want.  ABORT.\n"
        except (SamExceptions.SamException, CommandLineArgParser.InvalidCommand), ex:
            result = getTb(sys.exc_info())
        except:
            result = getTb(sys.exc_info())

        printFooter = 1
        return (printFooter, result)

    #---------------------------------------------------------------------

    def _invokeDumpster(self, dumpster, argdict):
        """
        Get the output from the appropriate dumpster.
        """
        now = time.strftime("%d %b %Y %H:%M:%S")

        dumpCmd = dumpster.getCommand()
        dumpCmd.setArgDict(argdict)
        cmdAsString = dumpCmd.getCommandAsString()
        try:
            dump = dumpster(**argdict) 
        except:
            dump = str(sys.exc_info()[1])


        # fix non-html characters:
        for (old, new) in [ ('&', '&amp;'),
                            ('<', '&lt;'),
                            ('>', '&gt;') ]:
            dump = string.replace(dump, old, new)
        return  "\n$# %s\n$ %s\n%s" % (now, cmdAsString, dump)
    #---------------------------------------------------------------------

    def _parseForm(self, form):
        argdict = {}
        dumpster = None
        for k in form.keys():
            argdict.update( {k:form[k].value} )

        # backwards compatibility for older implementation:
        if argdict.has_key('serverType'):
            st = string.upper( argdict.get('serverType') )
            if st == 'STATION':
                dumpster = self.sam.dumpStation
                argdict.update( {'all':1} )
                if argdict.has_key('serverName'):
                    argdict.update( {'station': argdict.get('serverName')} )
                elif argdict.has_key('stationName'):
                    argdict.update( {'station': argdict.get('stationName')} )
            elif st == 'FSS':
                dumpster = self.sam.dumpFss
                if argdict.has_key('serverName'):
                    argdict.update( {'station': argdict.get('serverName')} )
                elif argdict.has_key('stationName'):
                    argdict.update( {'station': argdict.get('stationName')} )
            elif st == 'PROJECT':
                dumpster = self.sam.dumpProject
                if argdict.has_key('serverName'):
                    argdict.update( {'station': argdict.get('serverName')} )
                elif argdict.has_key('stationName'):
                    argdict.update( {'station': argdict.get('stationName')} )
                argdict.update( {'project':argdict.get('projectName')} )

            elif st == 'DBSERVER':
                dumpster = self.sam.dumpDbServer
                if argdict.has_key('serverName'):
                    argdict.update({'dbserver' : argdict.get('serverName')})
        else:

            # figure out what they want to dump
            if argdict.has_key('cid') or argdict.has_key('consumerId'):
                dumpster = self.sam.dumpConsumer
            elif argdict.has_key('project'):
                dumpster = self.sam.dumpProject
            elif argdict.has_key('file') or argdict.has_key('filename'):
                dumpster = self.sam.dumpFile
            elif argdict.has_key('fss'):
                dumpster = self.sam.dumpFss
                argdict.update( {'station':argdict.get('fss')} )
            elif argdict.has_key('station'):
                dumpster = self.sam.dumpStation
                argdict.update( {'all':TRUE_} )
            elif argdict.has_key('dbserver'):
                dumpster = self.sam.dumpDbServer
                if argdict.get('dbserver') is None or argdict.get('dbserver') == '':
                    argdict.update({'dbserver':os.environ.get('SAM_DB_SERVER_NAME')})

        # update with timeout parameters:
        argdict.update({ATTR_RETRY_MAX_COUNT: RETRY_MAX_COUNT,
                        ATTR_CORBA_CALL_TIMEOUT: CORBA_CALL_TIMEOUT})
        return (dumpster, argdict)
#================================================================================

if __name__ == "__main__":
    from SamUtility import FastCgiHandler
    
    TITLE_         = FastCgiHandler.TITLE_
    CONTENT_TYPE_  = FastCgiHandler.CONTENT_TYPE_
    PAGE_HEADER_   = FastCgiHandler.PAGE_HEADER_
    BG_COLOR_      = FastCgiHandler.BG_COLOR_
    PRE_FORMATTED_ = FastCgiHandler.PRE_FORMATTED_
    PRINT_COUNTER_ = FastCgiHandler.PRINT_COUNTER_

    # were we called as a FastCgiScript, or as a python script?
    calledAs = sys.argv[0]
    if string.find(calledAs, ".py") >= 0:
        printCounter = FALSE_
    else:
        printCounter = TRUE_

    processing = sys.argv[0]

    kwargs = {TITLE_         : 'sam dump',
              BG_COLOR_      : "#ffff99",
              PRE_FORMATTED_ : TRUE_,
              PRINT_COUNTER_ : printCounter }

    fcgi = FastCgiHandler.FastCgiHandler(DumpSam, **kwargs)

    #
    # fcgi will invoke:
    #    thing = DumpSam(kwargs)
    #    (printFooter, outString) = thing.handle(form, outdevice)
    #    if printFooter:
    #        print(outString)
    #        
    # Class to be called must implement handle method,
    #  which returns a tuple  (printFooter, outString).
    # printFooter:  flag, set this to TRUE if you want the
    #  fcgi handler to print the outString and the closing </body></html></etc>;
    # set it to FALSE if this does not need to be done (i.e., either
    #  done in the handler process, or the handler has forked off subprocesses
    #  and this is the return from one of the now-dead subprocesses, which
    #  you do not want to process.
    #
    fcgi.handle()
