#!/bin/sh

prog=`basename $0`

#Edit the lines below to accomodate your local settings
#increase chances of success calling Condor commands
PATH=/opt/condor/bin/:$PATH
unset CONDOR_CONFIG

#And now, read through the end of the file to make sure
#,This will work for YOUR particaular Condor installation

die () { echo "$prog: $@" >&2; exit 1; }
get_value () { echo $1 | sed "s?^--[a-z|-]*=??g"; }
usage () { die Possible actions: `egrep '^  [a-z].*\)$' $0 | sed 's?)??g'`; }

#Arguments:
action=$1

shift
[ -n "$action" ] || usage
while [ -n "$1" ]; do
  case $1 in
  --executable=*)
    executable=`get_value "$1"`
    ;;

  --arguments=*)
    arguments=`get_value "$1"`
    ;;

  --project=*)
    id=`get_value "$1"`
    ;;

  --stdout=*)
    stdout=`get_value "$1"`
    ;;

  --stderr=*)
    stderr=`get_value "$1"`
    ;;

  --queue=*)
    queue=`get_value "$1"`
    ;;

  --local-job-id=*)
    local_job_id=`get_value "$1"`
    ;;

  --require-no-preemption=*)
    requireNoPreemption=`get_value "$1"`
    ;;
  *)
    die "Invalid option: $1"
    usage
  esac
  shift
done

# Derived local variables
user_script_and_args="$executable $arguments"


#move multiple submissions to a level above until
# sam batch adapters have a template for it
if [ -z "$JIM_KLUDGY_NUM_JOBS" ]; then
 submitTimes=1
else
 submitTimes=$JIM_KLUDGY_NUM_JOBS
fi


get_ids () 
{
   # In condor commands even if job is not found return status is 0 (unlike pbs)
   out=`condor_q  -constraint "grid_jid == \"$1\"" -format '%d ' ClusterId`
	if [ $? != 0 ];then
      echo "FOUL"
   else
      echo "$out" 
   fi 	  
}

get_id_and_status_list () 
{
   gId=$1
   lId=${2:-""}
   out=`condor_q $lId -constraint "grid_jid == \"$1\"" -format '%d' ClusterId -format ' %d\n' JobStatus`
   if [ $? != 0 ];then
      echo "FOUL"
   else
      echo "$out" 
   fi 	  
}

convert_id_and_status_list () {
  awk 'BEGIN {
    JIM_JOB_SUBMITTED="submitted"
    JIM_JOB_PENDING="pending"
    JIM_JOB_ACTIVE="active"
    JIM_JOB_FAILED="failed"
    JIM_JOB_DONE="done"
    JIM_JOB_SUSPENDED="suspended"
    JIM_JOB_UNSUBMITTED="unsubmitted"
  }
  { 
    id=$1;
    condor_status=$2; /*1 = Idle, 2 = Running, 3 = Removed, 4 = Completed, and 5 = Held.*/
    if   (condor_status==1) {jim_job_status=JIM_JOB_PENDING}
    else if (condor_status==2) {jim_job_status=JIM_JOB_ACTIVE}
    else if (condor_status==3) {jim_job_status=JIM_JOB_DONE}
    else if (condor_status==4) {jim_job_status=JIM_JOB_DONE}
    else if (condor_status==5) {jim_job_status=JIM_JOB_SUSPENDED}
    else {jim_job_status="UNKNOWN"}
    printf("JobId=%s Status=%s\n",id,jim_job_status)
  }'

}


submit_batch_job() {
   echo "Submiting $job_type through batch ..."
   echo "Submiting $job_type $submitTimes times through batch ..."
   echo "Using condor_submit from `which condor_submit`..."

cat << EOF >> sam_condor_submit.${job_type}.$$.jdf 
universe = vanilla
should_transfer_files = YES
transfer_output_files = output_kludge
when_to_transfer_output = ON_EXIT_OR_EVICT
output = ${local_stdout_file}.${job_type}.\$(cluster).\$(process)
error = ${local_stderr_file}.${job_type}.\$(cluster).\$(process)
log = log.\$(cluster).\$(process)
executable = ${executable}
# JIM__LOCAL_JID needs to be just the cluster value
environment = SAM_STATION=$SAM_STATION;JIM_LOCAL_JID=\$(cluster);JIM_JOB_NUMBER=$JIM_JOB_NUMBER
+sam_project = "${id}"
+grid_jid = "${id}"
+IsCommitted = True
+IsCommittedJob = True
EOF


if [ "${requireNoPreemption}" = "TRUE" ];then
   cat << EOF >> sam_condor_submit.${job_type}.$$.jdf
Requirements = ((Disk > 3000000) && (IsDedicated == TRUE) && (ClusterGeneration==3) && (machine != "p48.cs.wisc.edu"))
queue $submitTimes
EOF

else
   cat << EOF >> sam_condor_submit.${job_type}.$$.jdf
Requirements = (Disk > 3000000) && ((UidDomain == "glow.wisc.edu")||((IsDedicated == TRUE) && (ClusterGeneration == 3) && (machine != "p48.cs.wisc.edu")))
queue $submitTimes
EOF

fi

exec condor_submit sam_condor_submit.${job_type}.$$.jdf

}


submit_interactive_job() {
   job_type=$1
   echo "Submiting $job_type interactively ..."
     #Today the shipping of the output and cleanup is handled in the job manager *polling* script, when the job finishes.
     #To invoke it, we must exit *succesfully* (trap "exit 0" TERM) from the job manager *submit* script: we do that by
     #sending TERM to this script after submitting the consumer to the BS.
     #Should we trigger the output handling and cleanup also in the job manager submit in case of failure to avoid this
     #cumbersome mechanism ?
   cmd='trap "exit 0" TERM; export GRANDPARENT=$$; echo $$; ./${user_script_and_args} 1>${local_stdout_file}.${job_type}.$$ 2>${local_stderr_file}.${job_type}.$$ & wait $!'
   echo "Executing: ${cmd}"
   eval ${cmd}
   echo "Submiting $job_type interactively ... done"
}


do_submit() {
   mode=$1
   job_type=$2
   
   if [ "$mode" = "INTERACTIVE" ]
   then
      submit_interactive_job $job_type
   else
      submit_batch_job $job_type
   fi
}
   
override_sam_environment() {
   sam_environment_file=`ls -t1 sam_batch_environment.* | head -1`
   echo "sam environment file: ${sam_environment_file}"
   grep SAM_STATION ${sam_environment_file} > ${sam_environment_file}.new
   grep SAM_PROJECT ${sam_environment_file} >> ${sam_environment_file}.new
   echo "export GW_FTP_SERVER_URL=\"${GW_FTP_SERVER_URL}\"" \
        >> ${sam_environment_file}.new
   echo "export GW_FTP_SERVER_IDENTITY=\"${GW_FTP_SERVER_IDENTITY}\"" \
        >> ${sam_environment_file}.new
   echo "export PATH=\$PATH:." >> ${sam_environment_file}.new
   echo "execute() {" >> ${sam_environment_file}.new
   echo "   cmd=\"\$@\"" >> ${sam_environment_file}.new
   echo "   echo Executing \${cmd}" >> ${sam_environment_file}.new
   echo "   eval \${cmd}" >> ${sam_environment_file}.new
   echo "}" >> ${sam_environment_file}.new
   mv ${sam_environment_file}.new ${sam_environment_file}
}

get_file_names() {

	local_user_output_file=`echo $GW_CUSTOM_OUT_TEMPL | \
                awk -F / '{print $(NF);}' | \
                cut -d '*' -f1`.\$\(cluster\).\$\(process\)
	local_stdout_file=`echo $GW_STD_OUT_TEMPL | \
        	awk -F / '{print $(NF);}' | cut -d '*' -f1`
    	local_stderr_file=`echo $GW_STD_ERR_TEMPL | \
                awk -F / '{print $(NF);}' | cut -d '*' -f1`

	echo GW_CUSTOM_OUT_TEMPL $GW_CUSTOM_OUT_TEMPL 
    	echo GW_STD_OUT_TEMPL $GW_STD_OUT_TEMPL 
	echo GW_STD_ERR_TEMPL $GW_STD_ERR_TEMPL 

}

gridftp_environment_definition() {
        #Submission requires user_script_and_args supplied and assumes several
        #env variables, see below

    if [ -z "$GW_SANDBOX_WORK_DIR" ]; then
      export GW_SANDBOX_WORK_DIR=`hostname -f`:`pwd`
      echo "setting GW_SANDBOX_WORK_DIR=$GW_SANDBOX_WORK_DIR"
    fi

    for v in GW_SANDBOX_WORK_DIR GW_CUSTOM_OUT_TEMPL GW_STD_OUT_TEMPL \
        GW_STD_ERR_TEMPL; do
      [ -n "`printenv $v`" ] || die "Environment var. $v not set"
    done

    echo GW_SANDBOX_WORK_DIR $GW_SANDBOX_WORK_DIR
    get_file_names
    
    X509_USER_CERT=`sam_gsi_read_config SAM_GSI_GRIDFTP_X509_CERT`
    X509_USER_KEY=`sam_gsi_read_config SAM_GSI_GRIDFTP_X509_KEY`

    client_proxy=${X509_USER_PROXY}
    sandbox_mgr_file=grid_extractor
    #sandbox_mgr_file=sandbox_mgr
    sandbox_self_extractor=${sandbox_mgr_file}.tgz.exe
    sam_client="${JIM_SANDBOX_DIR}/etc/sam.tgz"
    gridftp_client=gridftp_client.tar.gz
    if [ -z "$GW_FTP_SERVER_URL" ]; then 
      echo "Getting GW_FTP_SERVER_URL ..."
      export GW_FTP_SERVER_URL="`jim_gridftpd_tool get_urls`"
    fi
    [ -n "$GW_FTP_SERVER_URL" ] || \
	die "ERROR: failed to get the GW_FTP_SERVER_URL"
     echo GW_FTP_SERVER_URL  $GW_FTP_SERVER_URL 
   
    if [ -z "$GW_FTP_SERVER_IDENTITY" ]; then
      echo "Getting GW_FTP_SERVER_IDENTITY ..."
      export GW_FTP_SERVER_IDENTITY="`grid-cert-info -subject -file ${X509_USER_CERT}`"
    fi
    [ -n "$GW_FTP_SERVER_IDENTITY" ] || \
       die "ERROR: failed to get the GW_FTP_SERVER_IDENTITY"
    echo GW_FTP_SERVER_IDENTITY $GW_FTP_SERVER_IDENTITY
}


# AN: create_packages packages
# Any changes to the packages variable is reflected outside the function too.   

create_packages(){
    # create_packages resolves the links if any.
    tempPack=""
    packagesWithLinks="`ls ${SANDBOX_WORK_DIR_PATH}/* | tr '\012' ' '`" 
    for pck in $packagesWithLinks
    do
      if [ -d "$pck" ]
      then 
        echo "$prog: create_packages: skipping directory $pck" >&2
      elif [ -L "$pck" ]
      then
	 tempPack="`ls -l $pck | awk '{print $11}'` $tempPack"
      else	 
	 tempPack="$tempPack $pck"
      fi
    done 
    echo "$tempPack"
}



##########################################################################
#                              MAIN STARTS HERE                         
##########################################################################


case "$action" in 
  consumer_submit)

    job_type="CONSUMER"

    gridftp_environment_definition

      #The following are the files for sam analysis
    SANDBOX_WORK_DIR_PATH=`echo ${GW_SANDBOX_WORK_DIR} | cut -d ":" -f 2`
    #create_packages packages # can take an additional argument like the path ${SANDBOX_WORK_DIR_PATH}/sam_batch_*
    #packages=`ls ${SANDBOX_WORK_DIR_PATH}/sam_batch_* | tr '\012' ' '`
    #packages="${JIM_SANDBOX_DIR}/etc/sam.tgz $packages"
    
    #the prepare_bootstrap_script transfer everything from the current workdir: it is important
    #to create the links here, but the list of packages is not used here anymore
    #IMPORTANT: NO ONE CREATES A LINK FOR sam.tgz ANYMORE
    create_packages
    
      #5th arg is a blank separated list of package tar balls the job depends upon
    X509_USER_PROXY=${client_proxy} prepare_bootstrap_script \
                                    ${sandbox_self_extractor} \
				    ${user_script_and_args}
    
    ####################################################################
    # PM: Where should the GW_FTP_SERVER_URL url be exported? Should it
    #     be in sam' environment or in the tmeplate? It is not part of
    #     sam but we use the sam's environment file for time being.
    #     Once we move to the batch-batch mode it will be more clear.
    # GGPM: Put arguments to the consumer script  links to user sandbox
    #       port of the gridftp
    ####################################################################
    echo "Initializing variables for consumer ... done"

    # Purging the sam generated environment file since it assumes ups at WN
    echo "Over riding SAM environment ..."
    override_sam_environment
    echo "Over riding SAM environment ... done"

    # Submit the consumer 
    echo "Submiting consumer ${id} ..."
    do_submit "BATCH" $job_type
    echo "Submiting consumer ${id} ... done"
    
    echo "Grand parent $GRANDPARENT"
    ps -ef | grep $GRANDPARENT
    kill $GRANDPARENT
    
  ;;

  consumer_lookup)
    out=`get_ids ${id}`
    if [ `echo "$out"| grep FOUL` ];then
       die "$prog: CONSUMER LOOKUP FAILED"
	 fi

    if [ "${out}x" = "x" ];then
       echo $JOB_DONE; exit 0
    else 
       echo $JOB_ACTIVE; exit 0
    fi
  ;;

  consumer_kill)
    out=`get_ids ${id}`
    if [ `echo "$out"| grep FOUL` ];then
       die "$prog: CONSUMER KILL FAILED"
	 else
	    condor_rm $out
	 fi	 
  ;;

  project_lookup)
    project_processes="`ps -u $USER -o pid,args | grep -v 'grep' | \
                        grep ${id} | awk '{ printf("%s",$1)}'`"
    echo ${project_processes} > /tmp/ppp.6565
    if [ "${project_processes}x" = "x" ]
    then 
       echo $JOB_DONE; exit 0
    else 
       echo $JOB_ACTIVE; exit 0
    fi
  ;;

  project_kill)
    # Stop all consumers first
    out=`get_ids ${id}`
    if [ `echo "$out"| grep FOUL` ];then
       die "$prog: PROJECT KILL FAILED"
	 fi
	 
    consumer_ids="$out"
    if [ "${consumer_ids}x" != "x" ]
    then 
       condor_rm ${consumer_ids}
    fi
    # Stop the project
    cmd="sam stop project --force --project=${id}"
    eval $cmd
    echo
  ;;

  project_submit)

    
    #Submission requires user_script_and_args supplied and assumes several
    #env variables, see below
    echo "Initializing variables for project ..."

    local_user_output_file=`echo $GW_CUSTOM_OUT_TEMPL | \
                        awk -F / '{print $(NF);}' | \
                        cut -d '*' -f1`.\$\(cluster\).\$\(process\)
    local_stdout_file=`echo $GW_STD_OUT_TEMPL | \
                       awk -F / '{print $(NF);}' | cut -d '*' -f1`
    local_stderr_file=`echo $GW_STD_ERR_TEMPL | \
                       awk -F / '{print $(NF);}' | cut -d '*' -f1`
    job_type="PROJECT"
    echo "Initializing variables for project ... done"

    ###############################################################################
    #                                      IMPORTANT 
    # sam_batch_project_xxxx will get unchanged sam_batch_environment_xxxx 
    # sam_batch_consumer_xxxx will get changed sam_batch_environment_xxxx 
    ###############################################################################

    # Submit the project 
    do_submit "INTERACTIVE" $job_type
  ;;

  job_submit)

    job_type="JOB"
    get_file_names
    
    
      #The following are the files for sam analysis
#    SANDBOX_WORK_DIR_PATH=`echo ${GW_SANDBOX_WORK_DIR} | cut -d ":" -f 2`
    
#    cmd="X509_USER_PROXY=${client_proxy} prepare_bootstrap_script \
#                                    ${sandbox_self_extractor} \
#				    ${user_script_and_args}"
				    
#    echo "$prog: Creating the execution script via: $cmd"
#    eval $cmd || die "Failed to prepare submission"

    # Submit the consumer 
    do_submit "BATCH" $job_type; exit $?
    echo "U SHOULDNT BE SEEIN THIS MESSAGE"
  ;;

  job_lookup)
	 [ -z ${id} ] && die "Must provide the project id."
    localId=${local_job_id:-""}
    out=`get_id_and_status_list ${id} ${localId}`
    if [ `echo "$out"| grep FOUL` ];then
       die "$prog: JOB LOOKUP FAILED" 
	 else
	    echo "$out" | convert_id_and_status_list
	 fi
    ;;


  job_kill)
    if [ -n "${local_job_id}" ];then
       condor_rm ${local_job_id}
       exit $?
    fi
    [ -z ${id} ] && die "Must provide the project id."
    out=`get_ids ${id}`
    if [ `echo "$out"| grep FOUL` ];then
       die "$prog: JOB KILL FAILED"
	 else
	    condor_rm $out
	 fi	 
  ;;

  *)
    usage
esac

