#!/usr/bin/env sampy
################################################################################
# Project   : SAM
# Package : sam_cp
# $Id: sam_cp.py,v
# 1.29 2001/06/04 21:39:09 lauri Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import string
import socket
import os.path
import pwd

ENSTORE_FLAG_STRING = "enstore"
VIRTUAL_DOMAIN_LIST_ = [ "enstore" ]

class SamCpFileParser:
    #
    # _input is in the form [[station:]machine:][/path/to/a/file]
    # alternative input in the from [[station:]schema://domain:][context]
   
    localDomain = socket.getfqdn()
    localHostName = socket.gethostname()

    def __init__(self, aFileString=""):
        self._input = aFileString
        self._station = ""
        self._domain = ""
        self._dirname = ""
        self._basename = ""
        self._path = ""
        self._fullyQualifiedPath = ""

        schemaContext = string.split(aFileString, "://",1)
     
        fields = []

        if ( len(schemaContext) == 2 ):
           schemaParts = string.split(schemaContext[0],":",1)

           if ( len(schemaParts) == 2 ):
              fields.append(schemaParts[0]) # station
              schema = schemaParts[1];
           else:
              schema = schemaParts[0];
  
           schema += "://" 

           pos = string.find(schemaContext[1],":")

           fields.append(schema+schemaContext[1][:pos])
           fields.append(schemaContext[1][pos+1:])
        else:
           fields = string.split(aFileString, ":", 2)

        if ( len(fields) == 1 ):
            # aFileString = "/path/to/file"
            self._domain = self.localDomain
            self._path = aFileString

        elif ( len(fields) == 2 ):
            # aFileString = "machine.domain:/path/to/file"
            self._domain = fields[0]
            self._path = fields[1]

        elif ( len(fields) == 3 ):
            # aFileString = "station:machine.domain:/path/to/file"
            self._station = fields[0]
            self._domain = fields[1]
            self._path = fields[2]

        try:
            self._dirname = os.path.dirname(self._path)
            if ( self._dirname == "" ):
                self._dirname = "./"
                self._path = "./" + self._path
            self._basename = os.path.basename(self._path)
        except:
            pass

        self._fullyQualifiedPath = "%s:%s" % (self._domain, self._path)

    def station(self):
        return self._station

    def domain(self):
        return self._domain

    def path(self):
        return self._path

    def basename(self):
        return self._basename

    def dirname(self):
        return self._dirname

    def fullyQualifiedPath(self):
        return self._fullyQualifiedPath

    def parsedFileName(self):
        if ( self.isLocal() ):
            returnValue = self._path
        else:
            returnValue = self._fullyQualifiedPath
        return "%s" % returnValue

    def __str__(self):
        return self._input

    def isEnstore(self):
        return ( string.find(self.fullyQualifiedPath(), ENSTORE_FLAG_STRING ) != -1 )

    def isLocal(self):
        return ( self._domain == self.localDomain or
                 self._domain == self.localHostName )

    def isVirtualDomain(self):
        return ( VIRTUAL_DOMAIN_LIST_.count(self._domain) )

    

if __name__ == "__main__":

    for testFileString in ( "central-analysis:d0mino.fnal.gov:/scratch/boo/hiss",
                            "enstore-stken:/pnfs/sam/boo/",
                            "/path/to/the/righteous",
                            "filenameHere",
                            "d0ora1.fnal.gov:filenameHere",
                            "d0mino.fnal.gov:",
                            "d0ora3.fnal.gov:",
                            "sam",
                            "d0test:d0test.fnal.gov:/tmp/",
                            "d0olb.fnal.gov:/tmp/",
                            "lion:hpss://in2p3:cchpssd0:/d0",
                            ""
                            ):
        tst = SamCpFileParser(testFileString)
        print("inputString = %s" % testFileString)
        print("\t station()       = %s" % tst.station())
        print("\t  domain()       = %s" % tst.domain())
        print("\t    path()       = %s" % tst.path())
        print("\t dirname()       = %s" % tst.dirname())
        print("\tbasename()       = %s" % tst.basename())
        print("\tparsedFileName() = %s" % tst.parsedFileName())
        print("\tisLocal() = %s" % tst.isLocal())
        print("returnString = %s\n\n" % tst)
