#!/usr/bin/env sampy
################################################################################
# Project   : SAM
# Package : sam_cp
# $Id: sam_cp.py,v
# 1.29 2001/06/04 21:39:09 lauri Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import sys
import os
import re
import string


#==================================================
# BYPASS MAIN PROGRAM HERE:
#==================================================
def bypassMain(samCpUserScript, argList):
    #
    # here we are entirely bypassing the internal sam_cp logic.
    # we will call the script defined by samCpUserScript, and pass it the
    # args as given on the command line.
    #

    # we need to expand any environmental variables
    # in order for os.execvp to understand then:
    theCommand = os.path.expandvars(samCpUserScript)

    # the way os.execvp operates:
    # os.execvp(cmd, args)
    # where
    #   cmd is the path to the command
    #   args[0] is the C-style argv[0], i.e., the name
    #    of the command internally, NOT an actual 'argument' as seen by the user
    #   args[1:] are the command line args.
    args = ["%s" % os.path.basename(theCommand)]
    for arg in argList:
        args.append(os.path.expandvars(arg))

    try:
        print("sam_cp exec> %s %s" % (theCommand, string.join(args[1:])))
        # if the execvp succeeds, then the process context is switched
        #  to that process, and the caller will receive the return
        #  status of THAT process.
        os.execvp(theCommand, args)
    except Exception, ex:
        # but if we failed, for any reason, we gotta pass it on up.
        raise
        #raise SamCpFailure("%s %s" % (theCommand, string.join(args[1:])), sys.exc_info()[1])


if __name__ == "__main__":
    import SamCpWrapper
    samCpUserScript = os.environ.get(SamCpWrapper.SAM_CP_USER_SCRIPT_)
    if samCpUserScript is not None:
        sys.exit(bypassMain(samCpUserScript, sys.argv[1:]))
    else:
        sys.exit(SamCpWrapper.main(sys.argv[1:]))
