#!/usr/bin/env python
""" This python script is a xmldb wrapper for ID  Management """

# $source: idmgr.py,v$

import xmldb 
import sys 
import traceback 
from string import join,find
from time import gmtime,strftime,time,localtime

def getLocalDocName(gId,cId,lId):
   """Returns the name of the XML document where the local JID info are stored"""
   return "%s_%s" % (gId,lId)
  
def getCollectionFromURI(globalDocumentURI):
   """Given an unique identifier for a global JID, returns an XMLDB collection
      where the document can be stored
   """
   import os.path
   return "%s/%s" % (os.path.dirname(globalDocumentURI),"globalJobs1")

class idmgr(xmldb.xmldb):
   """ This class provides API to manage Job IDs"""
	
   __collection_="/db/jobs"
   __API_Name_="idmgr"
   __API_Version_="1.0"

#   class IdmgrException(Exception):
#      def __init__(self, value):
#         self.value = value

#      def __str__(self):
#         result = "<IdmgrException message = %s>" %self.value
#         return result
#   # end class IdmgrException 
	
	
   def __init__(self,url=None,globalDocumentURI=None):
      """This class creates a DB instance and
         Checks if jobs collection already exist
         Else create one & Create a new index
         based on Global Job Id"""

      if url:
         self.url = url
         xmldb.xmldb.__init__(self,self.url)
      else:
         xmldb.xmldb.__init__(self)
		
      if globalDocumentURI:
         self.__collection_ = getCollectionFromURI(globalDocumentURI)

      #Unmarshel collection String
      import string
      cols = string.split(self.__collection_,"/")
      cols.remove('')

      dbroot = cols[0]
      cols = cols[1:]
      
      colString = "/%s"  % dbroot

      for c in cols:
         if not self.IsCollection("%s/%s" %(colString,c)):
	    try:
	       self.CreateCollection(colString,c)
	       self.CreateIndexer("%s/%s" %(colString,c),"idiindex","global_job@Id")
	    except:
               print "Cannot create collection %s in Context %s" %(c,colString)
               traceback.print_exc(file=sys.stderr)
	       sys.exit(1)   
	       
         colString = "%s/%s" %(colString,c) 

   # end constructor for idmgr

   def getCurTime(self):
      return strftime("%b %d %Y, %H:%M:%S %Z", localtime())

   def getCurEpochTime(self):
      return time()

   def GetApiVersion(self):
      """ Returns API name & Version"""
      return "API Name = %s API Version = %s" \
              %(self.__API_Name_,self.__API_Version_);

   def GetConf(self):
      """Return the URL & Collection information"""
      return "URL = %s  Collection = %s"  \
              % (self.url,self.__collection_)


   def AddTimestamp (self,str):
      """Time stamp gets inserted to the root element
         of the XML structure"
         str - XML string
         returns the status of the operation"""

      # check if an attribute called Id exist in the root element of  str
      import re

      root_tag=re.search('<.*?>',str);

      # Replace the root element ">" (or) "\>" with creation time 
      # & ">" (or) "/>" respectively
      
      import string
      if (string.find(root_tag.group(0),"/>") == -1):
         str=string.replace (str,">"," created='%s' created_epoch='%s'>" \
                             % (self.getCurTime(), self.getCurEpochTime() ), 1)
      else:
         str=string.replace (str,">"," created='%s' created_epoch='%s'/>" \
                             % (self.getCurTime(), self.getCurEpochTime() ) , 1)
      return str




   def UpdateJob(self,action,DocName,select,str=""):
      """Update Job with XUpdate string
         DocName - the document to be updated
         If nothing is passed, then the XUpdate would be run against
         the collection.
         action - xmldb:Update operation
         select - XPath string
         str - XML string
         returns  the update results"""

      # XUpdate String
      modification= """<?xml version='1.0'?>
<xupdate:modifications version='1.0'
xmlns:xupdate='http://www.xmldb.org/xupdate'>  
<xupdate:%s select='%s'>%s</xupdate:%s > 
</xupdate:modifications>""" % (action,select,str,action)

      return self.UpdateDocument(self.__collection_,DocName,modification)
   ## end of UpdateJob ##

   def UpdateJobs(self,action,select,str=""):
      """Update Jobs with XUpdate string
      
         THIS METHOD IS OBSOLETE: IT IS NOW A NO-OP
         
         action - xmldb:Update operation
         select - XPath string
         str - XML string
         returns  the update results"""
      #xmldb:update String
      print >>sys.stderr, "WARNING: method UpdateJobs of idmgr package is obsolete "
      return """<?xml version="1.0"?>
<result count="0" />"""

   ## end of UpdateJobs ##

   def AddElement(self,Id,select,str):
      """Add an element to a node
         Id = Id of the node
         select = XPath expression
         str = XML string
         return the status of the opertation"""

      return self.UpdateJob("append",Id,select,str)
   ## end of AddElement


   def RemoveElement(self,Id,select):
      """Remove an element from a document
         gId = Id	
         select = XPath expression
         return the status of the opertation"""

      return self.UpdateJob("remove",Id,select)



   def QueryJob(self,Id,query):
      """Query Jobs with XPath string
         Query- XPath expression string
         Id - If Id is given, only that document will be queried
         If nothing is passed, then the query would be run against
         the collection.
         returns  the query results"""

      return self.QueryDocument(self.__collection_,Id,query)

   def QueryJobs(self,query):
      """Query Jobs with XPath string
         Query- XPath expression string
         returns  the query results"""

      return self.QueryCollection(self.__collection_,query)
   ## end of QueryJobs ##




   def CreateGlobalJob(self,gId,elements=None,owner=None,jobtype="Unknown",facility="Unknown"):
      """Create a new document with node "global_job"
         with attribues "Id" and "created"
         Optionally elements could be added"
         gId = Global Job Id
         elements = Optional node(s) that goes inside this global_job
	 owner = "Optional (owner of the job)
         returns status of the operation"""
	    
      #  Global Job  templete 
      #  "<global_job Id='%s'></global_job>" %(gId)

      if owner == None:	
         doc = "<global_job Id='%s' job_type='%s' facility='%s'></global_job>" % (gId,jobtype,facility)
      else :
         doc = "<global_job Id='%s' owner='%s' job_type='%s' facility='%s'></global_job>" % (gId,owner,jobtype,facility)

      doc = self.AddTimestamp(doc)
      result = self.InsertDocument(self.__collection_, gId,doc)

      #Update the Global Job documument with more entries
      if (elements != None):
         result = self.AddElement(gId,"/global_job",elements)
   
      return result 
   ## end of CreateGlobalJob ## 

   def CreateClusterJob(self,gId,cId,cName,elements=""):
      """Create a Cluster Job entry inside "/global_job"
         Optionally more elements could be added
         gId = Global Job Id
         cId = Cluster Job Id
         cName = Cluster Name
         elements = Optional node(s) that goes inside this cluster_job
         returns status of the operation"""

      doc = "<cluster_job Id='%s' cluster='%s'>%s</cluster_job>" \
               % (cId,cName,elements)
	    
      doc = self.AddTimestamp(doc)
	    
      # Update the Global Job documument with more entries
      return self.AddElement(gId,"/global_job",doc)
   ## end of CreateClusterJob ###   



   def CreateLocalJob(self,gId,cId,lId,elements=""):
      """Create a Local Job document which has
         gridId and clusterId as elements
         Optionally more elements could be added"
         gId = Global Job Id
         cId = Cluster Job Id
         lId = Local job Id
         elements = Optional node(s) that goes inside this local_job 
         returns status of the operation"""
    
      # local_job template
      doc = "<local_job Id='%s'><global_job>%s</global_job><cluster_job>\
             %s</cluster_job>%s</local_job>" % (lId,gId,cId,elements)
      doc = self.AddTimestamp(doc)
      insertedDoc = self.InsertDocument(self.__collection_, \
                                      getLocalDocName(gId,cId,lId), \
                                      doc)

      #Add local jid to global doc
      updateStr ='<local_job Id="%s"/>' % lId	    
      self.UpdateCluster('append',gId,cId,updateStr)

      return insertedDoc
      
      
      
   ## end of CreateLocalJob ##
   
         
   def GetGlobalJob(self,gId):
      """Get the global job document 
         gId = Global Id	
         returns Global Job Document"""
      return self.GetDocument(self.__collection_,gId)
   ## end of GetGlobalJob ##

   def GetClusterJob(self,gId,cId):
      """Returns the cluster element for a
         given cluster Id
         gId - Global Id
         cId - Cluster Id"""

      query = "/global_job[@Id=\"%s\"]/cluster_job[@Id=\"%s\"]" % (gId,cId)
      return self.QueryJob(gId,query)
   ## end of GetClusterJob ##



   def GetLocalJob(self,gId,cId,lId):
      """Returns the local job element for a
         given cluster Id and Global Id
         gId - Global Id - retained for backward compatibility
         cId - Cluster Id -retained for backward compatibility
         lId - Local Id"""

      return self.GetDocument(self.__collection_,getLocalDocName(gId,cId,lId))	    
   ## end of GetLocalJob ##      

   
   def GetClusterJobList(self,gId):
      """Returns a list of the cluster job ids for a
         given global Id
         gId - Global Id"""

      from xml_common_lib.PyXmlObject import PyXmlObject
      pyOb = PyXmlObject(self.GetGlobalJob(gId))
      i=0
      cIdList=[]
      while 1:
         try:
            cId=pyOb.global_job[0].cluster_job[i].Id
            cIdList.append(cId)
            i=i+1
         except:
            break

      return cIdList      
   ## end of GetClusterJobList ##


   def GetLocalJobList(self,gId,cId):
      """Returns a list of the local job ids for a
         given cluster Id and global Id
         gId - Global Id
         cId - Cluster Id"""

      query = "/global_job/cluster_job/local_job"
      localJobXMLResults = self.QueryDocument(self.__collection_, \
                                              gId,
                                              query)

      from xml_common_lib.PyXmlObject import PyXmlObject
      pyOb = PyXmlObject(localJobXMLResults)
      i=0
      lIdList=[]
      while 1:
         try:
            lId = pyOb.result[0].local_job[i].Id
            lIdList.append(lId)
            i=i+1
         except:
            break

      return lIdList
   ## end of GetLocalJobList ##


   def UpdateGlobal(self,action,gId,xml_str,path=""):
      """Action on a Global Job
         action - xmldb:update action
         gId- Global JobId
         xml_str - XML String to append
         path - Optional extra path  to append"""

      xpath = "/global_job%s" %(path)
      return self.UpdateJob(action,gId,xpath,xml_str)
   ## end of UpdateGlobal ##


   def UpdateCluster(self,action,gId,cId,xml_str,path=""):
      """Action on a Cluster Job
         action - xmldb:update action
         gid - Global JobId
         cId - Cluster JobId
         xml_str - XML string to append"""

      xpath="/global_job/cluster_job[@Id=\"%s\"]%s" %(cId,path)
      return self.UpdateJob(action,gId,xpath,xml_str)
   ## end of UpdateCluster ##


   def UpdateLocal(self,action,gId,cId,lId,xml_str,path=""):
      """Action on a Local Job
         action - xmldb:update action
         gid - Global JobId - retained for backward compatability
         cId - Cluster JobId- retained for backward compatability
         lid - Local JobId
         xml_str - XML string to append"""
   
      xpath="/local_job[@Id=\"%s\"]%s" %(lId,path)
      return self.UpdateJob(action,getLocalDocName(gId,cId,lId),xpath,xml_str)		
   ## end of UpdateLocal ##


   def UpdateLocalJobStatus(self,gId,cId,lId,st):
      """Adds the job status on a Local Job
         gid - Global JobId 
         cId - Cluster JobId
         lid - Local JobId
         st  - Status string (active, done,...)"""
      updateStr ='<status time=\"%s\" time_epoch=\"%s\" code=\"%s\"/>' % \
                  (self.getCurTime(), self.getCurEpochTime() ,st)	    
      return self.UpdateLocal('append',gId,cId,lId,updateStr)
   ## def UpdateLocalJobStatus ##


   def GetLatestLocalJobStatus(self,gId,cId,lId):
      """Returns a tuple containing 2 string:
         (the latest job status, its epoch time) 
         gId - Global Id 
         cId - Cluster Id
         cId - Local Id"""

      from xml_common_lib.PyXmlObject import PyXmlObject
      pyOb = PyXmlObject(self.GetLocalJob(gId,cId,lId))
      i=0
      latest_status=""
      latest_time=0
      while 1:
         try:
            status = pyOb.local_job[0].status[i].code
            time   = pyOb.local_job[0].status[i].time_epoch

            if time > latest_time:
               latest_status = status
               latest_time=time
         
	    i = i+1
         except:
            break
      
      return (latest_status, latest_time)	   
   ## end of GetLatestLocalJobStatus ##


   def UpdateLocalJobTargetedResource(self,gId,cId,lId,resource):
      """Adds the job status on a Local Job
         gid - Global JobId
         cId - Cluster JobId
         lid - Local JobId
         st  - Status string (active, done,...)"""
      updateStr ='<TargetedResource time=\"%s\" time_epoch=\"%s\" name=\"%s\"/>' % \
                  (self.getCurTime(), self.getCurEpochTime() ,resource)
      return self.UpdateLocal('append',gId,cId,lId,updateStr)
   ## def UpdateLocalJobTargetedResource ##


   def GetLatestLocalJobTargetedResource(self,gId,cId,lId):
      """Returns a tuple containing 2 string:
         (the latest job targeted resource, its epoch time) 
         gId - Global Id 
         cId - Cluster Id
         cId - Local Id"""

      from xml_common_lib.PyXmlObject import PyXmlObject
      pyOb = PyXmlObject(self.GetLocalJob(gId,cId,lId))
      i=0
      latest_resource=""
      latest_time=0
      while 1:
         try:
            resource = pyOb.local_job[0].TargetedResource[i].name
            time   = pyOb.local_job[0].TargetedResource[i].time_epoch

            if time > latest_time:
               latest_resource = resource
               latest_time=time
         
	    i = i+1
         except:
            break
      
      return (latest_resource, latest_time)	   
   ## end of GetLatestLocalJobTargetedResource ##



   def RemoveGlobalJob(self,gId):
      """Remove a Global Job document
         gId = Global Id	
         returns status of the operation"""
 		
      return self.RemoveDocument(self.__collection_,gId)
   ## end of RemoveGlobalJob ##      
      

   def RemoveClusterJob(self,gId,cId):
      """Removes a Cluster Job from a Global Job
         gId = Global Job Id
         cId = Cluster Job Id
         returns status of the operation"""
	 
      xpath_str = "/global_job/cluster_job[@Id=\"%s\"]" % (cId)	    
      return self.RemoveElement(gId,xpath_str)
   ## end of RemoveClusterJob ##       

   def RemoveLocalJob(self,gId,cId,lId):
      """Removes a Cluster Job from a Global Job
         gId = Global Job Id
         cId = Cluster Job Id
         lId = Local Job Id
         returns status of the operation"""
      return self.RemoveDocument(self.__collection_,getLocalDocName(gId,cId,lId))
   ## end of RemoveLocalJob ##







def test1():
  import os
  sep = "--------------------------------"
  print "Testing idmgr"
  gId="gridId"
  cId="clusterId"

  lId1="1234"
  lId2="5678"
  jobCollection="/db/jobs/GG_test"
  
  try:
    xmldbURL=os.environ["XMLDB_SERVER_URL"]
  except:
    raise "XML DB server url not known. Cannot talk to the database."

  jobIdMgr = idmgr(xmldbURL,jobCollection)	   
  
  print "CreateGlobalJob"
  print jobIdMgr.CreateGlobalJob(gId,"","Gabriele_Garzoglio")
  print "CreateClusterJob"
  print jobIdMgr.CreateClusterJob(gId,cId,"cluster_name")
  print "CreateLocalJob twice"
  print jobIdMgr.CreateLocalJob(gId,cId,lId1)
  print jobIdMgr.CreateLocalJob(gId,cId,lId2)

  print "GetGlobalJob=%s" % jobIdMgr.GetGlobalJob(gId)
  print "GetClusterJob=%s" % jobIdMgr.GetClusterJob(gId,cId)
  print "GetLocalJob 1=%s" % jobIdMgr.GetLocalJob(gId,cId,lId1)
  print "GetLocalJob 2=%s" % jobIdMgr.GetLocalJob(gId,cId,lId2)
  print "GetClusterJobList=%s" % jobIdMgr.GetClusterJobList(gId)
  print "GetLocalJobList=%s" % jobIdMgr.GetLocalJobList(gId,cId)
  
  print "UpdateGlobal=%s" % jobIdMgr.UpdateGlobal("append",gId,"<appe>global_append</appe>")
  print "UpdateCluster=%s" % jobIdMgr.UpdateCluster("append",gId,cId,"<appe>dasdas</appe>")
  print "UpdateLocal 1=%s" % jobIdMgr.UpdateLocal("append",gId,cId,lId1,"<files>pythia</files>")
  print "UpdateLocalJobStatus 1=%s" % jobIdMgr.UpdateLocalJobStatus(gId,cId,lId1,"ENDED")
  
  a,b = jobIdMgr.GetLatestLocalJobStatus(gId,cId,lId1)
  print "GetLatestLocalJobStatus 1=(%s,%s)" % (a,b)
  
  #print jobIdMgr.RemoveLocalJob(gId,cId,lId1)
  #print jobIdMgr.RemoveLocalJob(gId,cId,lId2)
  #print jobIdMgr.RemoveClusterJob(gId,cId)
  #print jobIdMgr.RemoveGlobalJob(gId)
  
if __name__ == '__main__':
    sys.exit(test1())
