"""
This little python script is a client wrapper for talking to a  Xindice
Server using XMLRPC"""

__author__="Vijay Murthi <murthi@fnal.gov>"
__date__= " Feb 07 2002"
__version__="1.0"

# $source: xmldb.py,v$

import xmlrpclib

class XUpdateBuilder:
    """This utility class builds an XUpdate modification string"""
    
    def convert_quote(self,s):
	"""Replace " with &quot; for a given string 
	
	s = String"""

	import string
	return string.replace(s,'"','&quot;')
	
    def build(self,operation,xpath,elements):
	"""Build a XUpdate modification string"""

	return """<?xml version='1.0'?>
<xupdate:modifications version='1.0'
 xmlns:xupdate='http://www.xmldb.org/xupdate'>
  <xupdate:%s select='%s' >%s</xupdate:%s>
  </xupdate:modifications>""" % \
	(operation,self.convert_quote(xpath),elements,operation)

class xmldb:
    """This class abstracts a Xindice Client
    
    Client that talks to Xindice Server 1.1"""

    class XmldbException(Exception):
    	def __init__(self, value):
	    self.value = value

	def __str__(self):
	    result = "<XmldbException message = %s>" %self.value
	    return result

    def __init__(self,url=None):
        """Constructor - Create a Xindice Client
        
        url - eg. http://samadams.fnal.gov:9080/Xindice
        returns Server instance"""

	if(url == None):
	    import os
	    url = os.environ.get('XMLDB_SERVER_URL')
	    if (url == None):
		raise self.XmldbException("Environment variable : XMLDB_SERVER_URL not set")

        self.url = url
        self.xdbServer =  xmlrpclib.Server(url)

    def CreateCollection(self,collection,name,filer=None):
        """ Create a Collection

        collection -  Collection Context
	name - Collection Name to be created
	filer - Specific filer. Default HashFiler
        return collection list"""

        result = self.xdbServer.run ({"message":"CreateCollection","collection":collection,"name":name,"filer":"HashFiler"})

        return result['result']

    def RemoveCollection(self,collection,name):
        """ Delete a Collection

        collection -  Collection Context
	name - Collection Name to be deleted
        return Status of operation"""

        result = self.xdbServer.run ({"message":"RemoveCollection","collection":collection,"name":name})

        return result['result']

    def ListCollections(self,collection):
        """ List of the Collections

        collection -  Collection Context
        return list of collections"""

        result = self.xdbServer.run ({"message":"ListCollections","collection":collection})
        return result['result']

    def IsCollection(self,collection):
	""" Returns positive if a collection exist

	collection - Collection Name
	return 1 if collection exist (or) 0 otherwise"""

	#Unpack collection String       
	from re import split
	splitStr = split('/',collection);
	strLen = len(splitStr);

	#Pack Individual Strings
	from string import join
	colContext=join(splitStr[0:strLen-1],"/");
	col=join(splitStr[strLen-1:strLen]);

	try:
	    cols = self.ListCollections(colContext);
	except Exception, e:
	    return 0

	return  col in cols

    def ListDocuments(self,collection):
        """ List of Documents 

        collection -  Collection context
        return list of Documents"""

        result = self.xdbServer.run ({"message":"ListDocuments","collection":collection})
        return result['result']

    def InsertDocument(self,collection,name,document):
        """ Insert document in a Collection

        collection -  Collection 
	name - Name (or) Key associated with the document
	document - Document String
        return Status of the operation"""

        result = self.xdbServer.run ({"message":"InsertDocument","collection":collection,"name":name,"document":document})
        return result['result']

    def RemoveDocument(self,collection,name):
        """ Delete document in a Collection

        collection -  Collection 
	name - Name (or) Key associated with the document
        return Status of the operation"""

        result = self.xdbServer.run ({"message":"RemoveDocument","collection":collection,"name":name})
        return result['result']

    def GetDocument(self,collection,name):
        """ Get a  document 

        collection -  collection 
	name - Name (or) Key associated with the document
        return Document string
	NOW deprecated !!! use GetResource instead"""

        #result = self.xdbServer.run ({"message":"GetDocument","collection":collection,"name":name})
        result = self.xdbServer.run ({"message":"GetResource","collection":collection,"name":name})
        return result['result']

    def GetResource(self,collection,name):
        """ Get a  document 

        collection -  collection 
	name - Name (or) Key associated with the document
        return Document string"""

        result = self.xdbServer.run ({"message":"GetResource","collection":collection,"name":name})
        return result['result']
    
    def IsDocument(self,collection,document):
	""" Returns positive if a document exist in the collection

	collection - Collection Name
	document - Name of the document
	return 1 if document exist (or) 0 otherwise"""

	docs = self.ListDocuments(collection);

	return  document in docs

    def Query(self,collection,type,query,name=None,namespaces={}):
        """ Query (or) Update a document (or) collection

        collection -  Collection
	type - XPath (or) XUpdate
	query - Query String for XPath type and Modification string for XUpdate
	name - Name (or) Key associated with the document; None - if query is for a collection
	namespaces - 
        return Query result"""

        if(name):
            name=str(name) #To make sure that Java treats name as String type

	if (name == None):
            result =  self.xdbServer.run({"message":"Query","collection":collection,"type":type,"query":query,"namespaces":namespaces})
	else:
           result =  self.xdbServer.run({"message":"Query","collection":collection,"name":name,"type":type,"query":query,"namespaces":namespaces})
	return result['result']

    def QueryCollection(self,collection,query):
        """ Query a collection

        collection -  Collection name
	query - XPath query
        return Query result"""

        result =  self.Query(collection,"XPath",query)
	return result

    def UpdateCollection(self,collection,mod_string):
        """ Update a collection

        collection -  Collection name
	mod_string - XUpdate Modification String
        return Update status"""

        result =  self.Query(collection,"XUpdate",mod_string)
	return result

    def QueryDocument(self,collection,name,query):
        """ Query a document

        collection -  Collection name
	name - Name (or) Key associated with the document
	query - XPath query
        return Query result"""

        result =  self.Query(collection,"XPath",query,name)
	return result


    def AppendToDocument(self,collection,name,Xpath,appendstring):
       """ Paeend arbitary xml string to a document

        collection   - Collection name
	Xpath        - XPath to parent tag 
	sppendstring - XML String to append
        return Update status"""
	
       #print "AppendToDocument INVOKED"  
       xupdateBuilder = XUpdateBuilder()
       modstring = xupdateBuilder.build("append",Xpath,appendstring)
       return self.UpdateDocument(collection,name,modstring)


    def UpdateDocument(self,collection,name,mod_string):
        """ Update a document

        collection -  Collection name
	name - Name (or) Key associated with the document
	mod_string - XUpdate Modification String
        return Update status"""

        result =  self.Query(collection,"XUpdate",mod_string,name)
	return result

    def GetApiVersion(self):
        """ Print the API Name and API Version of the software

        return API Name_Version"""

        result = self.xdbServer.run ({"message":"GetApiVersion"})
	return result['result']

    def GetServerVersion(self):
	""" Print the Server Version 

        result = self.xdbServer.run ({"message":"GetServerVersion"})
        return Server version"""

    def GetCollectionCount(self,collection):
	"""  Get the Collection count in a given Collection

        collection -  Collection Name
	return Collection count"""

	result = self.xdbServer.run ({"message":"GetCollectionCount","collection":collection})
	return result['result']

    def GetDocumentCount(self,collection):
	"""  Get the Document count in a given Collection

        collection -  Collection Name
	return Collection count"""

	result = self.xdbServer.run ({"message":"GetDocumentCount","collection":collection})
	return result['result']

    def CreateIndexer(self,collection,name,pattern):
	""" Create a Indexer for a Collection

        collection -  Collection Name
        name - Indexer name to be created
        pattern -
        Pattern	         Description
        ========         =============
        elem             The value of the named element
        elem@attr        The value of the attribute for the named element
        *		 The value of all elements
        *@attr           The value of the named attribute for all elements
        elem@*           The value of all attributes for the named element
        *@*		 The value of all attributes for all elements        
	return Operation status"""

	result = self.xdbServer.run ({"message":"CreateIndexer","collection":collection,"name":name,"pattern":pattern})
	return result['result']

    def RemoveIndexer(self,collection,name):
	""" Remove a Indexer for a Collection

        collection -  Collection Name
        name - Indexer name to be deleted
	return Operation status"""

	result = self.xdbServer.run ({"message":"RemoveIndexer","collection":collection,"name":name})
	return result['result']


    def ListIndexers(self,collection):
	"""  List the Indexers in a given Collection

        collection -  Collection Name
	return List of Indexers"""

	result = self.xdbServer.run ({"message":"ListIndexers","collection":collection})
	return result['result']

    def GetUrl(self):
        """ Retrieves current url of the XML DB server"""
        return self.url
