"""This python script is command line xmldb client """

__author__="Vijay Murthi <murthi@fnal.gov>"
__date__= " Feb 17 2002"
__version__="0.1"

# $source: idmgr.py,v$

from xmldb import xmldb
import sys


class xmldb_command(xmldb):
	""" This class provides API for command-line xmldb"""

	__API_Name_="xmldb_client"
	__API_Version_="0.1"

	class XmldbCommandException(Exception):
	    def __init__(self, value):
		self.value = value

	    def __str__(self):
		result = {'xmldb_cmd.Fault':"'<"+self.value +"/>'"}
		return result ['xmldb_cmd.Fault']

	def __init__(self):
		""" This class initialize variables"""

		#Set variable url as unavailable
		self.url = 1;

		#Server Commands
		self.svrcommands ={
		    #action : Method Invocation 
		    'ac' :[self.CreateCollection],
		    'dc' :[self.RemoveCollection],
		    'lc' :[self.ListCollections],
		    'ld' :[self.ListDocuments],
		    'ad' :[self.InsertDocument],
		    'dd' :[self.RemoveDocument],
		    'rd' :[self.GetDocument],
		    'qc' :[self.QueryCollection],
		    'qd' :[self.QueryDocument],
		    'uc' :[self.UpdateCollection],
		    'ud' :[self.UpdateDocument],
		    'id' :[self.AppendToDocument],
                    'ci' :[self.CreateIndexer],
                    'ri' :[self.RemoveIndexer],
		}	

		#Helper
		self.svrhelper ={
		     #action : [Action String, No of arguments, [valid options], Usage]
		    'ac' :["AddCollection",["-c","-n","-url"], \
			    "xmldb_cmd ac -c /db -n test -url http://localhost:9080/Xindice"],
                    'dc' :["DeleteCollection",["-c","-n","-url"], \
			    "xmldb_cmd dc -c /db -n test -url http://localhost:9080/Xindice"],
                    'lc' :["ListCollections",["-c","-url"], \
			    "xmldb_cmd lc -c /db -url http://localhost:9080/Xindice"],
                    'ld' :["ListDocuments",["-c","-url"],\
			    "xmldb_cmd ld -c /db -url http://localhost:9080/Xindice"],
                    'ad' :["AddDocument",["-c","-n","-s","-url"], \
			    "xmldb_cmd ad -c /db/test -n sf [-s <modification_string> | -f <filename> ] -url http://localhost:9080/Xindice"],
                    'dd' :["DeleteDocument",["-c","-n","-url"],\
			    "xmldb_cmd dd -c /db/test -n sf -url http://localhost:9080/Xindice"],
                    'rd' :["ReadDocument",["-c","-n","-url"], \
			    "xmldb_cmd rd -c /db/test -n sf -url http://localhost:9080/Xindice"],
                    'qc' :["QueryCollection",["-c","-q","-url"], \
			    "xmldb_cmd qc -c /db/test -q /address -url http://localhost:9080/Xindice"],
                    'qd' :["QueryDocument",["-c","-n","-q","-url"], \
			    "xmldb_cmd qd -c /db/test -n sf -q /address -url http://localhost:9080/Xindice"],
                    'uc' :["UpdateCollection",["-c","-s","-url"], \
			    "xmldb_cmd uc -c /db/test [-s <modification_string> | -f <filename> ] -url http://localhost:9080/Xindice"],
                    'ud' :["UpdateDocument",["-c","-n","-s","-url"], \
			    "xmldb_cmd ud -c /db/test -n sf [-s <modification_string> | -f <filename>] -url http://localhost:9080/Xindice"],
                    'id' :["InsertIntoDocument",["-c","-n","-x","-xs","-url"], \
		            "xmldb_cmd id -c /db/test -x <XPath to parent tag> -xs <xml string to insert> -url http://localhost:9080/Xindice"],
                    'ci' :["CreateIndexer", [ "-c", "-n", "-p", "-url" ], \
                            """xmldb_cmd ci -c /db/test -n index_name  -p "*" -url http://localhost:9080/Xindice"""],
                    'ri' :["RemoveIndexer", [ "-c", "-n", "-url" ], \
                            """xmldb_cmd ri -c /db/test -n index_name   -url http://localhost:9080/Xindice"""],

                }

		#Client commands
		self.clicommands ={
		    #Option : Method Invocation
		    '--help' :self.PrintHelp,
		    '--version' :self.PrintApiVersion
		}

	def PrintDefault(self):
	    """ Default print script"""

	    self.PrintApiVersion();
	    print "Options:"
	    print "    --help [command]"
	    print "    --version"

	def PrintHelp(self,element=None):
	    """Print Help for commands"""

	    if (element == None):
		print "Usage: xmldb_cmd --help [command]"
		print "commmand         operations"
		for name,value in self.svrhelper.items():
		    print " %s       ----->  %s" %(name,value[0])
	    else:
		try:
		    print " Usage: %s  \n   %s" %(element,self.svrhelper[element][2])
		    print "\nOptions:"
		    for value in self.svrhelper[element][1]:
			if(value == "-c"):
			    print "%s        Collection Context" %value
			elif(value == "-n"):
			    print "%s        Name (or) Key " %value
			elif(value == "-s"):
			    print "%s | -f   String|File " %value
			elif(value == "-q"):
			    print "%s        Query" %value
			elif(value =="-url"):
			    print "%s      URL (Optional if XMLDB_SERVER_URL is set)" %value
			elif(value =="-x"):
			    print "%s        XPath to document" %value    
			elif(value =="-xs"):
			    print "%s       XML String" %value    
		except KeyError, e:
		    print "Command %s invalid" %element
		    self.PrintHelp()

	def PrintApiVersion(self):
	    """ Print the API Version and API Name"""

	    print "%s API Version = %s" \
	             %(self.__API_Name_,self.__API_Version_);

	def isURL(self):
	    """Check if a URL is provided by the user"""

	    return self.url;

	def setURL(self,url):
	    """Set if url is provided by user"""

	    self.url=url;
	

	def CheckOptions(self,elements):
	    """If the options are valid"""

	    self.dict={}
	    for i in (range(2,len(elements[2:])+2,2)):
		if (elements[i] == "-f"):
		    #If input is a file
		    f=open(elements[i+1],'r')
		    self.dict['-s'] = f.read()[:-1]
		else:
		    self.dict[elements[i]] = elements[i+1]

	    for value in self.svrhelper[elements[1]][1]:
		if(self.dict.has_key(value) ==0):
		   if (value == "-url"):
		    # Set url is FALSE
		    self.setURL(0); 
		   else:
		    sys.stderr.write("Option %s not found\n " %(value))
		    return 1;
	    return 0;

	def InvokeApi(self,command):
	    """If the options are valid"""
	    if (self.isURL()):
		server = xmldb.__init__(self,self.dict['-url'])
	    else:
		server = xmldb.__init__(self)

	    list = []
	    for element in self.svrhelper[command][1]:
		if (element != "-url"):
		    list.append(self.dict[element])

	    return (self.svrcommands[command][0])(*list)
